from __future__ import print_function
from builtins import (str, super)
from past.builtins import basestring

#import csv
#import os

#from . import __file__

from . import mixin
from . import core

from . import CoordinateConversion
from . import Datum


## --------------------------------------------------------------------
## Map coordinate conversion names to the set of coordinates to which
## they apply
## --------------------------------------------------------------------
#_name_to_coordinates = {}
#_file = os.path.join(os.path.dirname(__file__),
#                     'etc/coordinate_reference/coordinates.txt')
#for x in csv.reader(open(_file, 'r'), delimiter=' ', skipinitialspace=True):
#    if not x or x[0] == '#':
#        continue
#    _name_to_coordinates[x[0]] = set(x[1:])
#
## --------------------------------------------------------------------
## Map coordinate conversion names to the set of coordinates to which
## they apply
## --------------------------------------------------------------------
#_datum_parameters = []
#_file = os.path.join(os.path.dirname(__file__),
#                     'etc/coordinate_reference/datum_parameters.txt')
#for x in csv.reader(open(_file, 'r'), delimiter=' ', skipinitialspace=True):
#    if not x or x[0] == '#':
#        continue
#    _datum_parameters.append(x[0])
#    
#_datum_parameters  = set(_datum_parameters)
##_datum_ancillaries = set()


class CoordinateReference(mixin.NetCDFVariable,
                          mixin.Container,
                          core.CoordinateReference):
    '''A coordinate reference construct of the CF data model. 

A coordinate reference construct relates the coordinate values of the
coordinate system to locations in a planetary reference frame.

The domain of a field construct may contain various coordinate
systems, each of which is constructed from a subset of the dimension
and auxiliary coordinate constructs. For example, the domain of a
four-dimensional field construct may contain horizontal (y-x),
vertical (z), and temporal (t) coordinate systems. There may be more
than one of each of these, if there is more than one coordinate
construct applying to a particular spatiotemporal dimension (for
example, there could be both latitude-longitude and y-x projection
coordinate systems). In general, a coordinate system may be
constructed implicitly from any subset of the coordinate constructs,
yet a coordinate construct does not need to be explicitly or
exclusively associated with any coordinate system.

A coordinate system of the field construct can be explicitly defined
by a coordinate reference construct which relates the coordinate
values of the coordinate system to locations in a planetary reference
frame and consists of the following:

* References to the dimension coordinate and auxiliary coordinate
  constructs that define the coordinate system to which the coordinate
  reference construct applies. Note that the coordinate values are not
  relevant to the coordinate reference construct, only their
  properties.

..

* A definition of a datum specifying the zeroes of the dimension and
  auxiliary coordinate constructs which define the coordinate
  system. The datum may be implied by the metadata of the referenced
  dimension and auxiliary coordinate constructs, or explicitly
  provided.

..

* A coordinate conversion, which defines a formula for converting
  coordinate values taken from the dimension or auxiliary coordinate
  constructs to a different coordinate system. A coordinate reference
  construct relates the coordinate values of the field to locations in
  a planetary reference frame.

.. versionadded:: 1.7

    '''

#    _name_to_coordinates = _name_to_coordinates
#    _datum_parameters    = _datum_parameters
#    _datum_ancillaries   = _datum_ancillaries

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance._CoordinateConversion = CoordinateConversion
        instance._Datum                = Datum
        return instance
    #--- End: def

    def __init__(self, coordinates=None, datum=None,
                 coordinate_conversion=None,
                 source=None, copy=True):
        '''**Initialization**

:Parameters:

    coordinates: sequence of `str`, optional
        Identify the related dimension and auxiliary coordinate
        constructs by their construct identifiers. Ignored if the
        *source* parameter is set.

        *Example:*
          ``coordinates=['dimensioncoordinate2']``

        *Example:*
          ``coordinates=('dimensioncoordinate0', 'dimensioncoordinate1')``

        The coordinates may also be set after initialisation with the
        `coordinates` and `set_coordinate` methods.

    datum: `Datum`, optional
        Set the datum component of the coordinate reference
        construct. Ignored if the *source* parameter is set.

        The datum may also be set after initialisation with the
        `set_datum` method.

    coordinate_conversion: `CoordinateConversion`, optional
        Set the coordinate conversion component of the coordinate
        reference construct. Ignored if the *source* parameter is set.

        The coordinate conversion may also be set after initialisation
        with the `set_coordinate_conversion` method.

    source: optional
        Initialize the coordinates, datum and coordinate conversion
        from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy arguments prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__(
            coordinates=coordinates,
            datum=datum,
            coordinate_conversion=coordinate_conversion,
            source=source,
            copy=copy)
        
        self._initialise_netcdf(source)
    #--- End: def
   
    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

.. versionadded:: 1.7

        '''    
        return self.name(default=self.nc_get_variable(''))
    #--- End: def

    def dump(self, display=True, _omit_properties=None, field=None,
             key='', _level=0, _title=None, _construct_names=None,
             _auxiliary_coordinates=None, _dimension_coordinates=None):
        '''A full description of the coordinate reference construct.

Returns a description of all properties, including those of
components.

.. versionadded:: 1.7

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    out: `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''          
        indent0 = '    ' * _level
        indent1 = '    ' * (_level+1)
        indent2 = '    ' * (_level+2)

        if _title is None:
            string = ['{0}Coordinate Reference: {1}'.format(
                indent0, self.name(default=''))]
        else:
            string = [indent0 + _title]

        # Coordinate conversion parameter-valued terms
        coordinate_conversion = self.get_coordinate_conversion()
        for term, value in sorted(coordinate_conversion.parameters().items()):
            string.append("{0}Coordinate conversion:{1} = {2}".format(
                indent1, term, value))

        # Coordinate conversion domain ancillary-valued terms
#        if field:
#            for term, key in sorted(coordinate_conversion.domain_ancillaries().items()):
#                value = field.domain_ancillaries().get(key)
#                if value is not None:
#                    value = 'Domain Ancillary: '+value.name(default=key)
#                else:
#                    value = ''
#                string.append('{0}Coordinate conversion:{1} = {2}'.format(
#                    indent1, term, str(value)))
        if _construct_names:
            for term, key in sorted(coordinate_conversion.domain_ancillaries().items()):
#                value = field.domain_ancillaries().get(key)
#                if value is not None:
#                    value = 'Domain Ancillary: '+value.name(default=key)
#                else:
#                    value = ''
                string.append('{0}Coordinate conversion:{1} = Domain Ancillary: {2}'.format(
                    indent1, term, _construct_names.get(key, 'cid%{}'.format(key))))
        else:
            for term, value in sorted(coordinate_conversion.ancillaries.items()):
                string.append("{0}Coordinate conversion:{1} = {2}".format(
                    indent1, term, str(value)))

        # Datum parameter-valued terms
        datum = self.get_datum()
        for term, value in sorted(datum.parameters().items()):
            string.append("{0}Datum:{1} = {2}".format(indent1, term, value))

#        # Datum domain ancillary-valued terms
#        if field:
#            for term, key in sorted(datum.domain_ancillaries().items()):
#                value = field.domain_ancillaries().get(key)
#                if value is not None:
#                    value = 'Domain Ancillary: '+value.name(default=key)
#                else:
#                    value = ''
#                string.append('{0}Datum:{1} = {2}'.format(indent1, term, str(value)))
#        else:
#            for term, value in sorted(datum.ancillaries().items()):
#                string.append("{0}Datum:{1} = {2}".format(indent1, term, str(value)))

        # Coordinates 
#       if field:
#           for key in sorted(self.coordinates()):
#               coord = field.coordinates().get(key)
#               if coord is not None:
#                   if isinstance(coord, DimensionCoordinate):
#                       coord = 'Dimension Coordinate: '+coord.name(default='key%'+key)
#                   elif isinstance(coord, AuxiliaryCoordinate):
#                       coord = 'Auxiliary Coordinate: '+coord.name(default='key%'+key)
#                   else:
#                       coord = coord.name(default=key)
#
#                   string.append('{0}{1}'.format(indent1, coord))
        if _construct_names:
            for key in sorted(self.coordinates(), reverse=True):
                coord = '{}'.format(_construct_names.get(key, 'cid%{}'.format(key)))
                if key in _dimension_coordinates:
                    coord = 'Dimension Coordinate: '+coord
                elif key in _auxiliary_coordinates:
                    coord = 'Auxiliary Coordinate: '+coord
                    
                string.append('{0}{1}'.format(indent1, coord))
        else:
            for identifier in sorted(self.coordinates()):
                string.append('{0}{1}'.format(indent1, identifier))
            
        string = '\n'.join(string)
       
        if display:
            print(string)
        else:
            return string
    #--- End: def
            
    def equals(self, other, rtol=None, atol=None, traceback=False,
               ignore_type=False):
        '''Whether two coordinate reference constructs are the same.

Equality is strict by default. This means that for two coordinate
reference constructs to be considered equal:

* the datum and coordinate conversion components must have the same
  string and numerical parameters.

The dimension coordinate, auxiliary coordinate and domain ancillary
constructs of the coordinate reference constructs are *not*
considered, because they may ony be correctly interpreted by the field
constructs that contain the coordinate reference constructs in
question. They are, however, taken into account when two fields
constructs are tested for equality.

Two numerical parameters ``a`` and ``b`` are considered equal if
``|a-b|<=atol+rtol|b|``, where ``atol`` (the tolerance on absolute
differences) and ``rtol`` (the tolerance on relative differences) are
positive, typically very small numbers. The data type of the numbers
is not taken into consideration. See the *atol* and *rtol* parameters.

Any type of object may be tested but, in general, equality is only
possible with another coordinate reference construct, or a subclass of
one. See the *ignore_type* parameter.

NetCDF elements, such as netCDF variable and dimension names, do not
constitute part of the CF data model and so are not checked.

.. versionadded:: 1.7

:Parameters:

    other: 
        The object to compare for equality.

    atol: float, optional
        The tolerance on absolute differences between real
        numbers. The default value is set by the `cfdm.ATOL` function.
        
    rtol: float, optional
        The tolerance on relative differences between real
        numbers. The default value is set by the `cfdm.RTOL` function.

    traceback: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_type: `bool`, optional
        Any type of object may be tested but, in general, equality is
        only possible with another coordinate reference construct, or
        a subclass of one. If *ignore_type* is True then then
        ``CoordinateReference(source=other)`` is tested, rather than
        the ``other`` defined by the *other* parameter.

:Returns: 
  
    out: `bool`
        Whether the two coordinate reference constructs are equal.

**Examples:**

>>> c.equals(c)
True
>>> c.equals(c.copy())
True
>>> c.equals('not a coordinate reference')
False

        '''
        pp = super()._equals_preprocess(other, traceback=traceback,
                                        ignore_type=ignore_type)
        if pp in (True, False):
            return pp
        
        other = pp
        
#        # Check for object identity
#        if self is other:
#            return True
#
#        # Check that each object is of the same type
#        if ignore_type:
#            if not isinstance(other, self.__class__):
#                other = type(self)(source=other, copy=False)
#        elif not isinstance(other, self.__class__):
#            if traceback:
#                print("{0}: Incompatible types: {0}, {1}".format(
#		    self.__class__.__name__,
#		    other.__class__.__name__))
#            return False
        
#        if not super().equals(
#                other, #rtol=rtol, atol=atol,
#                traceback=traceback,
#                ignore_type=ignore_type):
#            return False

        coords0 = self.coordinates()
        coords1 = other.coordinates()
        if len(coords0) != len(coords1):
            if traceback:
                print(
"{}: Different sized collections of coordinates ({}, {})".format(
    self.__class__.__name__, coords0, coords1))
            return False

        if not self.coordinate_conversion.equals(
                other.coordinate_conversion,
                rtol=rtol, atol=atol,
                traceback=traceback,
                ignore_type=ignore_type):
            if traceback:
                print(
"{}: Different coordinate conversions".format(self.__class__.__name__))
            return False
        
        if not self.datum.equals(
                other.datum,
                rtol=rtol, atol=atol,
                traceback=traceback,
                ignore_type=ignore_type):
            if traceback:
                print(
"{}: Different datums".format(self.__class__.__name__))
            return False

        # Still here? Then the two coordinate references are as equal
        # as can be ascertained in the absence of domains.
        return True
    #--- End: def

    def name(self, default=None, ncvar=True, custom=None,
             all_names=False):
        '''Return a name.

By default the name is the first found of the following:

  1. The `standard_name` CF property.
  
  2. The `!id` attribute.
  
  3. The `long_name` CF property, preceeded by the string
     ``'long_name:'``.
  
  4. The `!ncvar` attribute, preceeded by the string ``'ncvar%'``.
  
  5. The value of the *default* parameter.

Note that ``f.name(identity=True)`` is equivalent to ``f.identity()``.

.. seealso:: `identity`

**Examples**

>>> n = r.name()
>>> n = r.name(default='NO NAME'))
'''
#        n = self.coordinate_conversion.get_parameter('standard_name', None)
#        if n is not None:
#            return n
#        
#        n = self.coordinate_conversion.get_parameter('grid_mapping_name', None)
#        if n is not None:
#            return n
#        
#        return default


        out = []

        if custom is None:
            custom = ('standard_name', 'grid_mapping_name')
        elif isinstance(custom, basestring):
            custom = (custom,)
            
        for prop in custom:
            n = self.coordinate_conversion.get_parameter(prop, None)
            if n is not None:
                out.append(str(n))
                if not all_names:
                    break
        #--- End: if
        
        if ncvar and (all_names or not out):
            n = self.nc_get_variable(None)
            if n is not None:
                out.append('ncvar%{0}'.format(n))
        #--- End: if

        if all_names:
            return out
        
        if out:
            return out[-1]

        return default

    
    #--- End: def

#    def _parse_match(self, match):
#        '''Called by `match`
#
#:Parameters:
#
#    match: 
#        As for the *match* parameter of `match` method.
#
#:Returns:
#
#    out: `list`
#        '''        
#        if not match:
#            return ()
#
#        if not isinstance(match, (list, tuple)): #basestring, dict, Query)):
#            match = (match,)
#
#        matches = []
#        for m in match:            
#            if isinstance(m, basestring):
#                if ':' in m:
#                    # CF property (string-valued)
#                    m = m.split(':')
#                    matches.append({m[0]: ':'.join(m[1:])})
#                else:
#                    # Identity (string-valued) or python attribute
#                    # (string-valued) or axis type
#                    matches.append({None: m})
#
#            elif isinstance(m, dict):
#                # Dictionary
#                matches.append(m)
#
#            else:
#                # Identity (not string-valued, e.g. cf.Query).
#                matches.append({None: m})
#        #--- End: for
#
#        return matches
#    #--- End: def
#
#    def match(self, description=None, inverse=False):
#        '''Test whether or not the coordinate reference satisfies the given
#conditions.
#
#:Returns:
#
#    out: `bool`
#        True if the coordinate reference satisfies the given criteria,
#        False otherwise.
#
#**Examples**
#
#        '''
##        conditions_have_been_set = False
##        something_has_matched    = False
#
#        description = self._parse_match(description)
#
##        if description:
##            conditions_have_been_set = True
#
#        found_match = True
#        for match in description:
#            found_match = True
#            
#            for prop, value in match.iteritems():
#                if prop is None: 
#                    if isinstance(value, basestring):
#                        if value in ('T', 'X', 'Y', 'Z'):
#                            # Axis type, e.g. 'T'
#                            x = getattr(v, value, False)
#                            value = True
#                        else:
#                            y = value.split('%')
#                            if len(y) > 1:
#                                # String-valued python attribute,
#                                # e.g. 'ncvar%latlon'
#                                x = getattr(v, y[0], None)
#                                value = '%'.join(y[1:])
#                            else:
#                                # String-valued identity
#                                x = v.identity(default=None)
#                    else:   
#                        # Non-string-valued identity
#                        x = v.identity(default=None)
#                else:
#                    x = v.get(prop)
#
#                if x is None:
#                    found_match = False
#                elif value is None:
#                    found_match = True
#                else:
#                    found_match = (value == x)
#                    try:
#                        found_match == True
#                    except ValueError:
#                        found_match = False
#                #--- End: if
#
#                if not found_match:
#                    break
#            #--- End: for
#
#            if found_match:
#                break
#        #--- End: for
#
#        return not bool(inverse)
#    #--- End: def

#    @classmethod
#    def _name_to_coordinates
#    # Map coordinate conversion names to their
#    _name_to_coordinates = _name_to_coordinates
    
#--- End: class



from builtins import object
from future.utils import with_metaclass

import abc


class ConstructAccess(with_metaclass(abc.ABCMeta, object)):
    '''Mixin class for accessing an embedded `Constructs` object.

    '''   
    @abc.abstractmethod
    def _get_constructs(self, *default):
        '''Return the `Constructs` object

:Parameters:

    default: optional
        If set then return *default* if there is no `Constructs` object.

:Returns:

    out:
        The `Constructs` object. If unset then return *default* if provided.

**Examples:**

>>> c = f._get_constructs()
>>> c = f._get_constructs(None)

        '''
        raise NotImplementedError()
    #--- End: def
    
    def array_constructs(self, copy=False):
        return self._get_constructs().array_constructs(copy=copy)
 
    def construct_axes(self, cid=None):
        '''Return the domain axes spanned by metadata construct data arrays.

.. versionadded:: 1.7

.. seealso:: `constructs`, `get_construct`

:Parameters:

    cid: `str`
        The identifier of the construct.

:Returns:

    out: `tuple` or `None`

        The identifiers of the domain axes constructs spanned by data
        array of metadata constructs. If a metadata construct does not have a
        data array then `None` is returned.

**Examples:**

>>> f.construct_axes('auxiliarycoordinate0')
('domainaxis1', 'domainaxis0')
>>> print(f.construct_axes('auxiliarycoordinate99'))
None

        '''
        return self._get_constructs().construct_axes(cid=cid)
    #--- End: def
    
    def construct_type(self, cid):
        '''TODO
        '''                
        return self._get_constructs().construct_type(cid)
    #--- End: def
    
    def constructs(self, construct_type=None, copy=False):
        '''Return metadata constructs.

.. versionadded:: 1.7

.. seealso:: `construct_axes`

:Parameters:

    construct_type: TODO

    copy: `bool`, optional
        If `True` then deep copies of the constructs are returned.

:Returns:

    out: `dict`
        TODO

**Examples:**

TODO

        '''
        return self._get_constructs().constructs(construct_type=construct_type,
                                                 copy=copy)
    #--- End: def
    
    @abc.abstractmethod
    def del_construct(self, cid):
        '''Remove a construct.

.. versionadded:: 1.7

.. seealso:: `get_construct`, `constructs`

:Parameters:

    cid: `str`, optional
        The identifier of the construct.

        *Example:*
          ``cid='auxiliarycoordinate0'``
        
:Returns:

    out: 
        The removed construct.

**Examples:**

TODO

        '''
        raise NotImplementedError()
    #--- End: def

    def get_construct(self, cid):
        '''Return a metadata construct.

.. versionadded:: 1.7

:Parameters:

    cid: `str`
        TODO

:Returns:

    out:
        TODO

**Examples:**

>>> f.constructs()
>>> f.get_construct('dimensioncoordinate1')
<>
>>> f.get_construct('dimensioncoordinate99', 'Not set')
'Not set'
        '''
        return self._get_constructs().get_construct(cid)
    #--- End: def

    def get_construct_axes(self, cid, *default):
        '''Return the domain axes spanned by a metadata construct data array.

.. versionadded:: 1.7

.. seealso:: `construct_axes`, `get_construct`, `set_construct_axes`

:Parameters:

    cid: `str`
        The construct identifier of the metadata construct.

        *Example:*
          ``cid='domainancillary1'``

    default: optional
        Return *default* if the metadata construct does not have a
        data array, or no domain axes have been set.

:Returns:

    out:
        The identifiers of the domain axis constructs spanned by data
        array. If a metadata construct does not have a data array, or
        no domain axes have been set, then `None` is returned.

**Examples:**

>>> f.get_construct_axes('domainancillary2')
('domainaxis1', 'domainaxis0')
>>> da = f.get_construct('domainancillary2')
>>> data = da.del_data()
>>> print(f.get_construct_axes('domainancillary2', None))
None

>>> f.get_construct_axes('cellmethod0', 'no axes')
'no axes'

        '''
        axes = self.construct_axes().get(cid)
        if axes is None:
            if default:
                return default[0]

            raise ValueError("no axes.")

        return axes
    #--- End: def
    
    def has_construct(self, cid):
        '''Whether a construct exisits.

.. versionadded:: 1.7

:Parameters:

    cid: `str`
        TODO

:Returns:

    out: `bool`
        True if the construct exists, otherwise False.

**Examples:**

TODO
        '''
        return self._get_constructs().has_construct(cid)
    #--- End: def

    def domain_axis_name(self, axis):
        '''TODO WHY DO WE NED THIS HERE?
        '''
        return self._get_constructs().domain_axis_name(axis)
    #--- End: def
    
    def set_construct(self, construct, cid=None, axes=None,
                      copy=True):
        '''Set a metadata construct.

.. versionadded:: 1.7

.. seealso:: `constructs`, `del_construct`, `get_construct`,
             `set_construct_axes`

:Parameters:

    construct:
        The metadata construct to be inserted.

    axes: sequence of `str`, optional
        The construct identifiers of the domain axis constructs
        spanned by the data array. An exception is raised if used for
        a metadata construct that can not have a data array,
        i.e. domain axis, cell method and coordinate reference
        constructs.

        The axes may also be set afterwards with the
        `set_construct_axes` method.

        *Example:*
          ``axes=['domainaxis1']``
        
        *Example:*
          ``axes=['domainaxis1', 'domainaxis0']``
        
    cid: `str`, optional
        The construct identifier to be used for the construct. If not
        set then a new, unique identifier is created automatically. If
        the identifier already exisits then the exisiting construct
        will be replaced.

        *Example:*
          ``cid='cellmeasure0'``

    copy: `bool`, optional
        If True then return a copy of the unique selected
        construct. By default the construct is not copied.
        
:Returns:

     out: `str`
        The construct identifier for the construct.
    
**Examples:**

>>> cid = f.set_construct(c)
>>> cid = f.set_construct(c, copy=False)
>>> cid = f.set_construct(c, axes=['domainaxis2'])
>>> cid = f.set_construct(c, cid='cellmeasure0')

        '''
        return self._get_constructs().set_construct(construct, cid=cid,
                                                    axes=axes,
#                                                    extra_axes=extra_axes,
#                                                    replace=replace,
                                                    copy=copy)
    #--- End: def

    def set_construct_axes(self, cid, axes):
        '''Set the domain axis constructs spanned by a metadata construct data
array.

.. versionadded:: 1.7

.. seealso:: `constructs`, `del_construct`, `get_construct`,
             `set_construct`

:Parameters:

    cid: `str`
        The construct identifier of the metadata construct.

        *Example:*
          ``cid='dimensioncoordinate2'``

     axes: sequence of `str`
        The construct identifiers of the domain axis constructs
        spanned by the data array.

        The axes may also be set with the `set_construct` method.

        *Example:*
          ``axes=['domainaxis1']``

        *Example:*
          ``axes=['domainaxis1', 'domainaxis0']``

:Returns:

    `None`

**Examples:**

>>> cid = f.set_construct(c)
>>> f.set_construct_axes(cid, axes=['domainaxis1'])

        '''
        return self._get_constructs().set_construct_axes(cid, axes)
    #--- End: def

#--- End: class



from future.utils import with_metaclass
from builtins import (range, super)

import abc

import numpy

from .array import Array


class CompressedArray(with_metaclass(abc.ABCMeta, Array)):
    '''Abstract base class for a container of an underlying compressed
array

See `cfdm.GatheredArray` for an example implementation.

    '''
    def __init__(self, compressed_array=None, shape=None, size=None,
                 ndim=None, compressed_dimension=None,
                 compression_type=None, **kwargs):
        '''**Initialization**

:Parameters:

    compressed_array: numpy array or subclass of `Array`
        The compressed array.

    shape: `tuple`
        The uncompressed array dimension sizes.

    size: `int`
        Number of elements in the uncompressed array.

    ndim: `int`
        The number of uncompressed array dimensions

    compressed_dimension: `int`
        The position of the compressed dimension in the compressed
        array.

    compression_type: `str`
        The type of compression.        
        
    kwargs: *optional*
        Further named parameters and their values needed to define the
        compressed array.

        '''
        super().__init__(shape=shape, size=size, ndim=ndim,
                         compressed_dimension=compressed_dimension,
                         compression_type=compression_type, **kwargs)

        self._set_compressed_Array(compressed_array)
    #--- End: def

    @abc.abstractmethod
    def __getitem__(self, indices):
        '''Return an uncompressed subspace as an independent numpy array.

x.__getitem__(indices) <==> x[indices]

The indices that define the subspace are relative to the uncompressed
array.

Indexing follows rules that are very similar to the numpy indexing
rules, the only differences being:

* An integer index i takes the i-th element but does not reduce the
  rank by one.

..

* When two or more dimensions' indices are sequences of integers then
  these indices work independently along each dimension (similar to
  the way vector subscripts work in Fortran). This is the same
  behaviour as indexing on a Variable object of the netCDF4 package.

.. versionadded:: 1.7

        '''
        raise NotImplementedError()
    #--- End: def

    def _get_compressed_Array(self, *default):
        '''TODO

:Parameters:

    default: *optional*
        If there is no data then *default* if returned if set.

:Returns:

    out:
        The data. If the data has not been set then *default* is
        returned, if set.

**Examples:**

>>> a = d.get_data(None)

        '''
        array = self._get_component('compressed_Array', None)

        if array is None:
            if default:
                return default[0]     

            raise AttributeError("{!r} has no data".format(
                self.__class__.__name__))
        
        return array   
    #--- End: def

    def _set_compressed_Array(self, array, copy=True):
        '''Set the compressed array.

.. versionadded:: 1.7

:Parameters:

    array: `numpy` array_like or subclass of `cfdm.data.Array`
        The compressed data to be inserted.

:Returns:

    `None`

:Examples:

>>> d._set_compressed_Array(a)

        '''
        if not isinstance(array, Array):
            raise TypeError("asdads  0000000000000000000")
#            if not isinstance(array, numpy.ndarray):
#                data = numpy.asanyarray(array)
#                
#            array = NumpyArray(array)

        if copy:
            array = array.copy()

        self._set_component('compressed_Array', array, copy=False)
    #--- End: def

    @property
    def dtype(self):
        '''Data-type of the data elements.

:Examples:

>>> a.dtype
dtype('float64')
>>> print(type(a.dtype))
<type 'numpy.dtype'>

        '''
        return self._get_compressed_Array().dtype
    #--- End: def

    @property
    def ndim(self):
        '''The number of dimensions of the uncompressed data.

:Examples:

>>> d.shape
(73, 96)
>>> d.ndim
2
>>> d.size
7008

>>> d.shape
(1, 1, 1)
>>> d.ndim
3
>>> d.size
1

>>> d.shape
()
>>> d.ndim
0
>>> d.size
1

        '''
        return self._get_component('ndim')
    #--- End: def

    @property
    def shape(self):
        '''Shape of the uncompressed data.

:Examples:

>>> d.shape
(73, 96)
>>> d.ndim
2
>>> d.size
7008

>>> d.shape
(1, 1, 1)
>>> d.ndim
3
>>> d.size
1

>>> d.shape
()
>>> d.ndim
0
>>> d.size
1

        '''
        return self._get_component('shape')
    #--- End: def

    @property
    def size(self):
        '''Number of elements in the uncompressed data.

:Examples:

>>> d.shape
(73, 96)
>>> d.size
7008
>>> d.ndim
2

>>> d.shape
(1, 1, 1)
>>> d.ndim
3
>>> d.size
1

>>> d.shape
()
>>> d.ndim
0
>>> d.size
1

        '''
        return self._get_component('size')
    #--- End: def

    def get_compressed_axes(self):
        '''Return axes that are compressed in the underlying array.

:Returns:

    out: `list`
        The compressed axes described by their integer positions.

:Examples:

>>> c.ndim
4
>>> c.compressed_array.ndim
3
>>> c.compressed_axes()
[1, 2]

        '''
        compressed_dimension = self.get_compressed_dimension()
        compressed_ndim = self._get_compressed_Array().ndim
        
        return list(range(compressed_dimension, self.ndim - (compressed_ndim - compressed_dimension - 1)))
    #--- End: def

    def get_array(self):
        '''Return an independent numpy array containing the uncompressed data.

:Returns:

    out: `numpy.ndarray`
        The uncompressed array.

:Examples:

>>> n = a.get_array()
>>> isinstance(n, numpy.ndarray)
True

        '''
        return self[...]
    #--- End: def

    def get_compressed_array(self):
        '''Return an independent numpy array containing the compressed data.

:Returns:

    out: `numpy.ndarray`
        The compressed array.

:Examples:

>>> n = a.get_compressed_array()
>>> isinstance(n, numpy.ndarray)
True

        '''
        return self._get_compressed_Array().get_array()
    #--- End: def

#--- End: class



from __future__ import print_function
from past.builtins import basestring

import os

#from .. import __version__

from .. import CF

from .. import (AuxiliaryCoordinate,
                CellMethod,
                CellMeasure,
                CoordinateReference,
                DimensionCoordinate,
                DomainAncillary,
                DomainAxis,
                Field,
                FieldAncillary,
                Bounds,
                Count,
                List,
                Index,
                CoordinateConversion,
                Datum)

from ..data import (Data,
                    GatheredArray,
                    NetCDFArray,
                    RaggedContiguousArray,
                    RaggedIndexedArray,
                    RaggedIndexedContiguousArray)

from . import CFDMImplementation

from .netcdf import NetCDFRead

implementation = CFDMImplementation(
    cf_version = CF(),
    
    AuxiliaryCoordinate = AuxiliaryCoordinate,
    CellMeasure         = CellMeasure,
    CellMethod          = CellMethod,
    CoordinateReference = CoordinateReference,
    DimensionCoordinate = DimensionCoordinate,
    DomainAncillary     = DomainAncillary,
    DomainAxis          = DomainAxis,
    Field               = Field,
    FieldAncillary      = FieldAncillary,
    
    Bounds = Bounds,
    List   = List,
    Index=Index,
    Count=Count,
    
    CoordinateConversion = CoordinateConversion,
    Datum                = Datum,
    
    Data                         = Data,
    GatheredArray                = GatheredArray,
    NetCDFArray                  = NetCDFArray,
    RaggedContiguousArray        = RaggedContiguousArray,
    RaggedIndexedArray           = RaggedIndexedArray,
    RaggedIndexedContiguousArray = RaggedIndexedContiguousArray,
)

def read(filename, external_files=None, create_field=None,
         verbose=False, _implementation=implementation):
    '''Read field constructs from a dataset.

The dataset may be a netCDF file on disk or on an OPeNDAP server.

The returned field constructs are sorted by the netCDF variable names
of their corresponding data variables.

**Performance**

Descriptive properties are always read into memory, but lazy loading
is employed for all data arrays, which means that no data is read into
memory until the data is required for inspection or to modify the
array contents. This maximises the number of field constructs that may
be read within a session, and makes the read operation fast.

.. versionadded:: 1.7

.. seealso:: `cfdm.write`, `cfdm.Field.create_field`,
             `cfdm.Field.read_report`

:Parameters:

    filename: `str`
        The file name or OPenDAP URL of the dataset.

        Relative paths are allowed, and standard tilde and shell
        parameter expansions are applied to the string.

        *Example:*
          The file ``file.nc`` in the user's home directory could be
          described by any of the following: ``'$HOME/file.nc'``,
          ``'${HOME}/file.nc'``, ``'~/file.nc'``,
          ``'~/tmp/../file.nc'``.
    
    external_files: (sequence of) `str`, optional
        Read external variables (i.e. variables which are named by
        attributes, but are not present, in the parent file given by
        the *filename* parameter) from the given external
        files. Ignored if the parent file does not contain a global
        "external_variables" attribute. Multiple external files may be
        provided, which are searched in random order for the required
        external variables.  
       
        If an external variable is not found in any external files, or
        is found in multiple external files, then the relevent
        metadata construct is still created, but without any metadata
        or data. In this case the construct's `!is_external` method
        will return `True`.

        *Example:*
          ``external_files='cell_measure.nc'``

        *Example:*
          ``external_files=['cell_measure_A.nc', 'cell_measure_O.nc']``

    create_field: (sequence of) `str`, optional
        Create extra, independent fields from the particular types of
        metadata constructs. The *create_field* parameter may be one,
        or a sequence, of:

          ==========================  ================================
          *create_field*              Metadata constructs
          ==========================  ================================
          ``'field_ancillary'``       Field ancillary constructs
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ==========================  ================================

        *Example:*
          To create fields from auxiliary coordinate constructs:
          ``create_field='auxiliary_coordinate'`` or
          ``create_field=['auxiliary_coordinate']``.

        *Example:*
          To create fields from domain ancillary and cell measure
          constructs: ``create_field=['domain_ancillary',
          'cell_measure']``.

        An extra field construct created via the *create_field*
        parameter will have a domain limited to that which can be
        inferred from the corresponding netCDF variable, but without
        the connections that are defined by the parent netCDF data
        variable. It is possible to create independent fields from
        metadata constructs that do incorporate as much of the parent
        field construct's domain as possible by using the
        `~cfdm.Field.create_field` method of a returned field
        construct, instead of setting the *create_field* parameter.

:Returns:
    
    out: `list`
        The field constructs found in the dataset. The list may be
        empty.

**Examples:**

TODO

    '''
    # Parse the field parameter
    if create_field is None:
        create_field = ()
    elif isinstance(create_field, basestring):
        create_field = (create_field,)

    filename = os.path.expanduser(os.path.expandvars(filename))
    
    if os.path.isdir(filename):
        raise IOError("Can't read directory {}".format(filename))

    if not os.path.isfile(filename):
        raise IOError("Can't read non-existent file {}".format(filename))

    # ----------------------------------------------------------------
    # Read the fields in the file
    # ----------------------------------------------------------------
    return _read_a_file(filename,
                        external_files=external_files,
                        create_field=create_field,
                        verbose=verbose,
                        _implementation=_implementation)
#--- End: def

def _read_a_file(filename,
                 external_files=(),
                 create_field=(),
                 verbose=False,
                 _implementation=None):
    '''Read the contents of a single file into a field list.

:Parameters:

    filename: `str`
        The file name.
    
:Returns:

    out: `list`
        The fields in the file.

    '''
    # ----------------------------------------------------------------
    # Initialise a netCDF read object
    # ----------------------------------------------------------------
    netcdf = NetCDFRead(_implementation)

    # ----------------------------------------------------------------
    # Read the file into fields.
    # ----------------------------------------------------------------
    if netcdf.is_netcdf_file(filename):
        fields = netcdf.read(filename, external_files=external_files,
                             create_field=create_field,
                             verbose=verbose)
    else:
        raise IOError("Can't determine format of file {}".format(filename))

    # ----------------------------------------------------------------
    # Return the fields
    # ----------------------------------------------------------------
    return fields
#--- End: def



from . import abstract

from ..core.data import NumpyArray as core_NumpyArray


class NumpyArray(abstract.Array, core_NumpyArray):
    '''An underlying numpy array.

    '''
    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

Returns a subspace of the array as an independent numpy array.

The indices that define the subspace must be either `Ellipsis` or a
sequence that contains an index for each dimension. In the latter
case, each dimension's index must either be a `slice` object or a
sequence of two or more integers.

Indexing is similar to numpy indexing. The only difference to numpy
indexing (given the restrictions on the type of indices allowed) is:

  * When two or more dimension's indices are sequences of integers
    then these indices work independently along each dimension
    (similar to the way vector subscripts work in Fortran).

        '''
        return self.get_subspace(self._get_component('array'), indices,
                                 copy=True)
    #--- End: def

#--- End: class



from builtins import (object, super)


class Compressed(object):
    '''TODO
    '''
    def _set_compressed_Array(self, array, copy=True):
        '''Set the compressed array.

.. versionadded:: 1.7

:Parameters:

    array: `numpy` array_like or subclass of `cfdm.data.Array`
        The compressed data to be inserted.

:Returns:

    `None`

:Examples:

>>> d._set_compressed_Array(a)

        '''
        if not isinstance(array, abstract.Array):
            if not isinstance(array, numpy.ndarray):
                data = numpy.asanyarray(array)
                
            array = NumpyArray(array)
            
        super()._set_compressed_Array(array, copy=copy)
    #--- End: def

#--- End: class



from __future__ import print_function
from builtins import (range, super)

from . import PropertiesData

from ..functions import RTOL, ATOL


class PropertiesDataBounds(PropertiesData):
    '''Mixin class for a data array with descriptive properties and cell
bounds.

    '''
    
    def __getitem__(self, indices):
        '''Return a subspace of the construct defined by indices

f.__getitem__(indices) <==> f[indices]

The new subspace contains the same properties and similar components
to the original construct, but the latter are also subspaced over
their corresponding axes.

Indexing follows rules that are very similar to the numpy indexing
rules, the only differences being:

* An integer index i takes the i-th element but does not reduce the
  rank by one.

* When two or more dimensions' indices are sequences of integers then
  these indices work independently along each dimension (similar to
  the way vector subscripts work in Fortran). This is the same
  behaviour as indexing on a Variable object of the netCDF4 package.

:Returns:

    out:
        The subspace of the construct.

**Examples:**

>>> f.data.shape
(1, 10, 9)
>>> f[:, :, 1].data.shape
(1, 10, 1)
>>> f[:, 0].data.shape
(1, 1, 9)
>>> f[..., 6:3:-1, 3:6].data.shape
(1, 3, 3)
>>> f[0, [2, 9], [4, 8]].data.shape
(1, 2, 2)
>>> f[0, :, -2].data.shape
(1, 10, 1)

        '''
        if indices is Ellipsis:
            return self.copy()

#        # Parse the index
        if not isinstance(indices, tuple):
            indices = (indices,)

#        indices = parse_indices(self.shape, indices)

        new = super().__getitem__(indices)
        
        data = self.get_data(None)

        if data is not None:
            new.set_data(data[indices], copy=False)

        # Subspace the bounds, if there are any.
        self_bounds = self.get_bounds(None)
        if self_bounds is not None:
            data = self_bounds.get_data(None)
            if data is not None:
                # There is a bounds array
                bounds_indices = list(indices)
                bounds_indices.append(Ellipsis)
                if data.ndim <= 1 and not self.has_geometry():
                    index = bounds_indices[0]
                    if isinstance(index, slice):
                        if index.step < 0:
                            # This scalar or 1-d variable has been
                            # reversed so reverse its bounds (as per
                            # 7.1 of the conventions)
                            bounds_indices.append(slice(None, None, -1))
                    elif data.size > 1 and index[-1] < index[0]:
                        # This 1-d variable has been reversed so
                        # reverse its bounds (as per 7.1 of the
                        # conventions)
                        bounds_indices.append(slice(None, None, -1))
                #--- End: if
#                    else:
#                        bounds_indices.append(slice(None))
#                else:
#                    bounds_indices.append(Ellipsis)
    
                new_bounds = new.get_bounds()
                new_bounds.set_data(data[tuple(bounds_indices)], copy=False)
        #--- End: if

        # Subspace the interior ring array, if there are one.
        interior_ring = self.get_interior_ring(None)
        if interior_ring is not None:
            data = interior_ring.get_data(None)
            if data is not None:
                new_interior_ring = new.get_interior_ring()
                new_interior_ring.set_data(data[indices], copy=False)
        #--- End: if

#       # Subspace the ancillary arrays, if there are any.
#       new_ancillaries = new.ancillaries()
#       if new_ancillaries:
#           ancillary_indices = tuple(indices) + (Ellipsis,)
#           
#           for name, ancillary in self.ancillaries.iteritems():
#               data = ancillary.get_data(None)
#               if data is None:
#                   continue
#
#               new_ancillary = new_ancillaries[name]
#               new_ancillary.set_data(data[ancillary_indices], copy=False)
#       #--- End: if

        # Return the new bounded variable
        return new
    #--- End: def

    def dump(self, display=True, field=None, key=None,
             _omit_properties=None, _prefix='', _title=None,
             _create_title=True, _level=0, _axes=None,
             _axis_names=None):
        '''TODO Return a string containing a full description of the instance.

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed, i.e. ``f.dump()`` is equivalent to
        ``print f.dump(display=False)``.

    omit: sequence of `str`, optional
        Omit the given CF properties from the description.

    _prefix: optional
        Ignored.

:Returns:

    out: `None` or `str`
        A string containing the description.

:Examples:

        '''
        # ------------------------------------------------------------
        # Properties and Data
        # ------------------------------------------------------------
        string = super().dump(display=False, field=field, key=key,
                              _omit_properties=_omit_properties,
                              _prefix=_prefix, _title=_title,
                              _create_title=_create_title,
                              _level=_level, _axes=_axes,
                              _axis_names=_axis_names)

        string = [string]
        
        # ------------------------------------------------------------
        # Bounds
        # ------------------------------------------------------------
        bounds = self.get_bounds(None)
        if bounds is not None:
            string.append(bounds.dump(display=False, field=field,
                                      key=key,
                                      _prefix=_prefix+'Bounds:',
                                      _create_title=False,
                                      _level=_level, _axes=_axes,
                                      _axis_names=_axis_names))

        # ------------------------------------------------------------
        # Geometry type
        # ------------------------------------------------------------
        geometry = self.get_geometry(None)
        if geometry is not None:
            indent1 = '    ' * (_level + 1)
            string.append(
                '{0}{1}Geometry: {2}'.format(indent1, _prefix, geometry))

        #-------------------------------------------------------------
        # Interior ring
        # ------------------------------------------------------------
        interior_ring = self.get_interior_ring(None)
        if interior_ring is not None:
            string.append(interior_ring.dump(display=False,
                                             field=field, key=key,
                                             _prefix=_prefix+'Interior ring:',
                                             _create_title=False,
                                             _level=_level,
                                             _axes=_axes,
                                             _axis_names=_axis_names))
            
        string = '\n'.join(string)
        
        if display:
            print(string)
        else:
            return string
    #--- End: def

    def equals(self, other, rtol=None, atol=None, traceback=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_properties=(), ignore_compression=False,
               ignore_type=False):
        '''Whether two data arrays with descriptive properties and cell bounds
are the same.

Equality is strict by default. This means that:

* the descriptive properties must be the same, and vector-valued
  properties must have same the size and be element-wise equal (see
  the *ignore_properties* parameter), and

..

* vector-valued properties must have same size and be element-wise
  equal,

..

* if there are data arrays then they must have same shape and data
  type, the same missing data mask, and be element-wise equal (see the
  *ignore_data_type* parameter).

..

* if there are bounds then their descriptive properties (if any) must
  be the same and their data arrays must have same shape and data
  type, the same missing data mask, and be element-wise equal (see the
  *ignore_properties* and *ignore_data_type* parameters).

Two numerical elements ``a`` and ``b`` are considered equal if
``|a-b|<=atol+rtol|b|``, where ``atol`` (the tolerance on absolute
differences) and ``rtol`` (the tolerance on relative differences) are
positive, typically very small numbers. See the *atol* and *rtol*
parameters.

If data arrays are compressed then the compression type and the
underlying compressed arrays must be the same, as well as the arrays
in their uncompressed forms. See the *ignore_compression* parameter.

Any type of object may be tested but, in general, equality is only
possible with another field construct, or a subclass of one. See the
*ignore_type* parameter.

NetCDF elements, such as netCDF variable and dimension names, do not
constitute part of the CF data model and so are not checked.

.. versionadded:: 1.7

:Parameters:


    other: 
        The object to compare for equality.

    atol: float, optional
        The tolerance on absolute differences between real
        numbers. The default value is set by the `cfdm.ATOL` function.
        
    rtol: float, optional
        The tolerance on relative differences between real
        numbers. The default value is set by the `cfdm.RTOL` function.

    ignore_fill_value: `bool`, optional
        If True then the "_FillValue" and "missing_value" properties
        are omitted from the comparison.

    traceback: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_properties: sequence of `str`, optional
        The names of properties to omit from the comparison.

    ignore_data_type: `bool`, optional
        If True then ignore the data types in all numerical data array
        comparisons. By default different numerical data types imply
        inequality, regardless of whether the elements are within the
        tolerance for equality.

    ignore_compression: `bool`, optional
        If True then any compression applied to the underlying arrays
        is ignored and only the uncompressed arrays are tested for
        equality. By default the compression type and, if appliciable,
        the underlying compressed arrays must be the same, as well as
        the arrays in their uncompressed forms

    ignore_type: `bool`, optional
         Any type of object may be tested but, in general, equality is
        only possible with another TODO, or a subclass of one. If
        *ignore_type* is True then then
        ``PropertiesDataBounds(source=other)`` is tested, rather than
        the ``other`` defined by the *other* parameter.

:Returns: 
  
    out: `bool`
        TODO

**Examples:**

>>> p.equals(p)
True
>>> p.equals(p.copy())
True
>>> p.equals('not a colection of properties')
False

        '''    
        # ------------------------------------------------------------
        # Check the properties and data
        # ------------------------------------------------------------
        if not super().equals(other, rtol=rtol, atol=atol,
                              traceback=traceback,
                              ignore_data_type=ignore_data_type,
                              ignore_fill_value=ignore_fill_value,
                              ignore_properties=ignore_properties,
                              ignore_type=ignore_type,
                              ignore_compression=ignore_compression):
            if traceback:
                print("???????/")
            return False
    
        # ------------------------------------------------------------
        # Check the geometry type
        # ------------------------------------------------------------
        if self.get_geometry(None) != other.get_geometry(None):
            if traceback:
                print(
"{0}: Different geometry types: {1}, {2}".format(
    self.__class__.__name__, self.get_geometry(None), other.get_geometry(None)))
            return False

        # ------------------------------------------------------------
        # Check the bounds 
        # ------------------------------------------------------------
        self_has_bounds = self.has_bounds()
        if self_has_bounds != other.has_bounds():
            if traceback:
                print("{0}: Different {1}".format(self.__class__.__name__, attr))
            return False
                
        if self_has_bounds:            
            if not self._equals(self.get_bounds(), other.get_bounds(),
                                rtol=rtol, atol=atol,
                                traceback=traceback,
                                ignore_data_type=ignore_data_type,
                                ignore_type=ignore_type,
                                ignore_fill_value=ignore_fill_value,
                                ignore_compression=ignore_compression):
                if traceback:
                    print("{0}: Different {1}".format(self.__class__.__name__, attr))
                return False
        #--- End: if

        # ------------------------------------------------------------
        # Check the interior ring
        # ------------------------------------------------------------
        self_has_interior_ring = self.has_interior_ring()
        if self_has_interior_ring != other.has_interior_ring():
            if traceback:
                print("{0}: Different {1}".format(self.__class__.__name__, attr))
            return False
                
        if self_has_interior_ring:            
            if not self._equals(self.get_interior_ring(), other.get_interior_ring(),
                                rtol=rtol, atol=atol,
                                traceback=traceback,
                                ignore_data_type=ignore_data_type,
                                ignore_type=ignore_type,
                                ignore_fill_value=ignore_fill_value,
                                ignore_compression=ignore_compression):
                if traceback:
                    print("{0}: Different {1}".format(self.__class__.__name__, attr))
                return False
        #--- End: if

        return True
    #--- End: def
    
    def del_part_ncdim(self):
        '''TODO
        '''        
        return self._del_component('part_ncdim')
    #--- End: def

    def expand_dims(self, position):
        '''Expand the shape of the data array.

Insert a new size 1 axis into the data array. A corresponding axis is
also inserted into the bounds data array, if present.

.. versionadded:: 1.7

.. seealso:: `squeeze`, `transpose`

:Parameters:

    position: `int`, optional
        Specify the position that the new axis will have in the data
        array. By default the new axis has position 0, the slowest
        varying position. Negative integers counting from the last
        position are allowed.

        *Example:*
          ``position=2``

        *Example:*
          ``position=-1``

:Returns:

    out:
        The new construct with expanded data axes.

**Examples:**

>>> f.data.shape
(19, 73, 96)
>>> f.expand_dims(position=3).data.shape
(96, 73, 19, 1)
>>> g = f.expand_dims(position=-1)
>>> g.data.shape
(19, 73, 1, 96)
>>> f.bounds.data.shape
(19, 73, 96, 4)
>>> g.bounds.data.shape
(19, 73, 1, 96, 4)

        '''
        position = self._parse_axes([position])[0]
        
        c = super().expand_dims(position)
        
        # ------------------------------------------------------------
        # Expand the dims of the bounds
        # ------------------------------------------------------------
        bounds = c.get_bounds(None)
        if bounds is not None:
            c.set_bounds(bounds.expand_dims(position), copy=False)

        # ------------------------------------------------------------
        # Expand the dims of the interior_ring
        # ------------------------------------------------------------
        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            c.set_interior_ring(interior_ring.expand_dims(position), copy=False)

        return c
    #--- End: def
    
    def get_node_ncdim(self, *default):
        '''TODO
        '''        
        return self._get_component('node_ncdim', *default)
    #--- End: def

    def get_part_ncdim(self, *default):
        '''TODO
        '''        
        return self._get_component('part_ncdim', *default)
    #--- End: def

    def has_part_ncdim(self):
        '''TODO
        '''        
        return self._has_component('part_ncdim')
    #--- End: def

    def set_node_ncdim(self, value):
        '''TODO Set the netCDF name of the dimension of a node coordinate variable.

        '''
        return self._set_component('node_ncdim', value)
    #--- End: def

    def set_part_ncdim(self, value):
        '''TODO Set the netCDF name of the dimension of the part_node_count
variable.

        '''
        return self._set_component('part_ncdim', value)
    #--- End: def

    def squeeze(self, axes=None):
        '''Remove size one axes from the data array.

By default all size one axes are removed, but particular size one axes
may be selected for removal. Corresponding axes are also removed from
the bounds data array, if present.

.. versionadded:: 1.7

.. seealso:: `expand_dims`, `transpose`

:Parameters:

    axes: (sequence of) `int`
        The positions of the size one axes to be removed. By default
        all size one axes are removed. Each axis is identified by its
        original integer position. Negative integers counting from the
        last position are allowed.

        *Example:*
          ``axes=0``

        *Example:*
          ``axes=-2``

        *Example:*
          ``axes=[2, 0]``

:Returns:

    out:
        The new construct with removed data axes.

**Examples:**

>>> f.data.shape
(1, 73, 1, 96)
>>> f.squeeze().data.shape
(73, 96)
>>> f.squeeze(0).data.shape
(73, 1, 96)
>>> g = f.squeeze([-3, 2])
>>> g.data.shape
(73, 96)
>>> f.bounds.data.shape
(1, 73, 1, 96, 4)
>>> g.data.shape
(73, 96, 4)

        '''
        axes = self._parse_axes(axes)

        c = super().squeeze(axes)

        # ------------------------------------------------------------
        # Squeeze the bounds
        # ------------------------------------------------------------
        bounds = c.get_bounds(None)
        if bounds is not None:
            c.set_bounds(bounds.squeeze(axes), copy=False)

        # ------------------------------------------------------------
        # Squeeze the interior_ring
        # ------------------------------------------------------------
        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            c.set_bounds(interior_ring.squeeze(axes), copy=False)

        return c
    #--- End: def
    
    def transpose(self, axes=None):
        '''Permute the axes of the data array.

Corresponding axes of the bounds data array, if present, are also
permuted.

Note that if i) the data array is two-dimensional, ii) the two axes
have been permuted, and iii) each cell has four bounds values; then
columns 1 and 3 (counting from 0) of the bounds axis are swapped to
preserve contiguity bounds in adjacent cells. See section 7.1 "Cell
Boundaries" of the CF conventions for details.

.. seealso:: `expand_dims`, `squeeze`

:Parameters:

    axes: (sequence of) `int`
        The new axis order. By default the order is reversed. Each
        axis in the new order is identified by its original integer
        position. Negative integers counting from the last position
        are allowed.

        *Example:*
          ``axes=[2, 0, 1]``

        *Example:*
          ``axes=[-1, 0, 1]``

:Returns:

    out: 
         The new construct with permuted data axes.

**Examples:**

>>> f.data.shape
(19, 73, 96)
>>> f.tranpose().data.shape
(96, 73, 19)
>>> g = f.tranpose([1, 0, 2])
>>> g.data.shape
(73, 19, 96)
>>> f.bounds.data.shape
(19, 73, 96, 4)
>>> g.bounds.data.shape
(73, 19, 96, 4)

        '''
        if axes is None:
            axes = list(range(ndim-1, -1, -1))
        else:
            axes = self._parse_axes(axes)

        c = super().transpose(axes)

        # ------------------------------------------------------------
        # Transpose the bounds
        # ------------------------------------------------------------        
        bounds = c.get_bounds(None)
        if bounds is not None:
            data = bounds.get_data(None)
            if data is not None:            
                b_axes = axes[:]
                b_axes.extend.extend(list(range(c.ndim, data.ndim)))
                
                bounds = bounds.transpose(b_axes)
                c.set_bounds(bounds, copy=False)
                
                if (c.ndim == 2 and data.ndim == 3 and data.shape[-1] == 4 and 
                    b_axes[0:2] == [1, 0]):
                    # Swap elements 1 and 3 of the trailing dimension
                    # so that the values are still contiguous (if they
                    # ever were). See section 7.1 of the CF
                    # conventions.
                    data[:, :, slice(1, 4, 2)] = data[:, :, slice(3, 0, -2)]
                    bounds.set_data(data, copy=False)
        #--- End: if

        a_axes = axes
        a_axes.append(-1)

        # ------------------------------------------------------------
        # Transpose the interior_ring
        # ------------------------------------------------------------
        interior_ring = c.get_interior_ring(None)
        if interior_ring is not None:
            interior_ring_axes = axes[:]
            interior_ring_axes.extend(list(range(c.ndim, interior_ring.ndim)))
            c.set_interior_ring(interior_ring.transpose(interior_ring_axes),
                                copy=False)

        return c
    #--- End: def

#--- End: class



from __future__ import print_function
import datetime
import os
import tempfile
import unittest

import numpy
import netCDF4

import cfdm

def _make_contiguous_file(filename):        
    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
    
    n.Conventions = 'CF-1.7'
    n.featureType = 'timeSeries'
    
    station = n.createDimension('station', 4)
    obs     = n.createDimension('obs'    , 24)
    name_strlen = n.createDimension('name_strlen', 8)
    
    lon = n.createVariable('lon', 'f8', ('station',))
    lon.standard_name = "longitude"
    lon.long_name = "station longitude"
    lon.units = "degrees_east"
    lon[...] = [-23, 0, 67, 178]
    
    lat = n.createVariable('lat', 'f8', ('station',))
    lat.standard_name = "latitude"
    lat.long_name = "station latitude" 
    lat.units = "degrees_north"
    lat[...] = [-9, 2, 34, 78]
    
    alt = n.createVariable('alt', 'f8', ('station',))
    alt.long_name = "vertical distance above the surface" 
    alt.standard_name = "height" 
    alt.units = "m"
    alt.positive = "up"
    alt.axis = "Z"
    alt[...] = [0.5, 12.6, 23.7, 345]
    
    station_name = n.createVariable('station_name', 'S1',
                                    ('station', 'name_strlen'))
    station_name.long_name = "station name"
    station_name.cf_role = "timeseries_id"
    station_name[...] = numpy.array([[x for x in 'station1'],
                                     [x for x in 'station2'],
                                     [x for x in 'station3'],
                                     [x for x in 'station4']])
    
    station_info = n.createVariable('station_info', 'i4', ('station',))
    station_info.long_name = "some kind of station info"
    station_info[...] = [-10, -9, -8, -7]
    
    row_size = n.createVariable('row_size', 'i4', ('station',))
    row_size.long_name = "number of observations for this station"
    row_size.sample_dimension = "obs"
    row_size[...] = [3, 7, 5, 9]
    
    time = n.createVariable('time', 'f8', ('obs',))
    time.standard_name = "time";
    time.long_name = "time of measurement" 
    time.units = "days since 1970-01-01 00:00:00"
    time[ 0: 3] = [-3, -2, -1]
    time[ 3:10] = [1, 2, 3, 4, 5, 6, 7]
    time[10:15] = [0.5, 1.5, 2.5, 3.5, 4.5]
    time[15:24] = range(-2, 7)
    
    humidity = n.createVariable('humidity', 'f8', ('obs',), fill_value=-999.9)
    humidity.standard_name = "specific_humidity"
    humidity.coordinates = "time lat lon alt station_name station_info"
    humidity[ 0: 3] = numpy.arange(0, 3)
    humidity[ 3:10] = numpy.arange(1, 71, 10)
    humidity[10:15] = numpy.arange(2, 502, 100)
    humidity[15:24] = numpy.arange(3, 9003, 1000)
    
    temp = n.createVariable('temp', 'f8', ('obs',), fill_value=-999.9)
    temp.standard_name = "air_temperature"
    temp.units = "Celsius"
    temp.coordinates = "time lat lon alt station_name station_info"
    temp[...] = humidity[...] + 273.15
    
    n.close()
    
    return filename
#--- End: def

def _make_indexed_file(filename):        
    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
    
    n.Conventions = 'CF-1.7'
    n.featureType = 'timeSeries'
    
    station = n.createDimension('station', 4)
    obs     = n.createDimension('obs'    , None)
    name_strlen = n.createDimension('name_strlen', 8)
    
    lon = n.createVariable('lon', 'f8', ('station',))
    lon.standard_name = "longitude"
    lon.long_name = "station longitude"
    lon.units = "degrees_east"
    lon[...] = [-23, 0, 67, 178]
    
    lat = n.createVariable('lat', 'f8', ('station',))
    lat.standard_name = "latitude"
    lat.long_name = "station latitude" 
    lat.units = "degrees_north"
    lat[...] = [-9, 2, 34, 78]
    
    alt = n.createVariable('alt', 'f8', ('station',))
    alt.long_name = "vertical distance above the surface" 
    alt.standard_name = "height" 
    alt.units = "m"
    alt.positive = "up"
    alt.axis = "Z"
    alt[...] = [0.5, 12.6, 23.7, 345]
    
    station_name = n.createVariable('station_name', 'S1',
                                    ('station', 'name_strlen'))
    station_name.long_name = "station name"
    station_name.cf_role = "timeseries_id"
    station_name[...] = numpy.array([[x for x in 'station1'],
                                     [x for x in 'station2'],
                                     [x for x in 'station3'],
                                     [x for x in 'station4']])
    
    station_info = n.createVariable('station_info', 'i4', ('station',))
    station_info.long_name = "some kind of station info"
    station_info[...] = [-10, -9, -8, -7]
    
    #row_size[...] = [3, 7, 5, 9]
    stationIndex = n.createVariable('stationIndex', 'i4', ('obs',))
    stationIndex.long_name = "which station this obs is for"
    stationIndex.instance_dimension= "station"
    stationIndex[...] = [3, 2, 1, 0, 2, 3, 3, 3, 1, 1, 0, 2, 
                         3, 1, 0, 1, 2, 3, 2, 3, 3, 3, 1, 1]
    
    t = [[-3, -2, -1],
         [1, 2, 3, 4, 5, 6, 7],
         [0.5, 1.5, 2.5, 3.5, 4.5],
         range(-2, 7)]
    
    time = n.createVariable('time', 'f8', ('obs',))
    time.standard_name = "time";
    time.long_name = "time of measurement" 
    time.units = "days since 1970-01-01 00:00:00"
    ssi = [0, 0, 0, 0]
    for i, si in enumerate(stationIndex[...]):
        time[i] = t[si][ssi[si]]
        ssi[si] += 1
        
    humidity = n.createVariable('humidity', 'f8', ('obs',), fill_value=-999.9)
    humidity.standard_name = "specific_humidity"
    humidity.coordinates = "time lat lon alt station_name station_info"
    
    h = [numpy.arange(0, 3),
         numpy.arange(1, 71, 10),
         numpy.arange(2, 502, 100),
         numpy.arange(3, 9003, 1000)]
    
    ssi = [0, 0, 0, 0]
    for i, si in enumerate(stationIndex[...]):
        humidity[i] = h[si][ssi[si]]
        ssi[si] += 1
        
    temp = n.createVariable('temp', 'f8', ('obs',), fill_value=-999.9)
    temp.standard_name = "air_temperature"
    temp.units = "Celsius"
    temp.coordinates = "time lat lon alt station_name station_info"
    temp[...] = humidity[...] + 273.15
    
    n.close()
    
    return filename
#--- End: def

def _make_indexed_contiguous_file(filename):        
    n = netCDF4.Dataset(filename, 'w', format='NETCDF3_CLASSIC')
    
    n.Conventions = 'CF-1.6'
    n.featureType = "timeSeriesProfile"
    
    station = n.createDimension('station', 3)  # 3 stations
    profile = n.createDimension('profile', 58) # 53 profiles spreadover 4 stations, each at a different time
    obs      = n.createDimension('obs'    , None) # 
    name_strlen = n.createDimension('name_strlen', 8)
    
    lon = n.createVariable('lon', 'f8', ('station',))
    lon.standard_name = "longitude"
    lon.long_name = "station longitude"
    lon.units = "degrees_east"
    lon[...] = [-23, 0, 67]
    
    lat = n.createVariable('lat', 'f8', ('station',))
    lat.standard_name = "latitude"
    lat.long_name = "station latitude" 
    lat.units = "degrees_north"
    lat[...] = [-9, 2, 34]
    
    alt = n.createVariable('alt', 'f8', ('station',))
    alt.long_name = "vertical distance above the surface" 
    alt.standard_name = "height" 
    alt.units = "m"
    alt.positive = "up"
    alt.axis = "Z"
    alt[...] = [0.5, 12.6, 23.7]
    
    station_name = n.createVariable('station_name', 'S1',
                                    ('station', 'name_strlen'))
    station_name.long_name = "station name"
    station_name.cf_role = "timeseries_id"
    station_name[...] = numpy.array([[x for x in 'station1'],
                                     [x for x in 'station2'],
                                     [x for x in 'station3']])
    
    profile = n.createVariable('profile', 'i4', ('profile'))
    profile.cf_role = "profile_id"
    profile[...] = numpy.arange(58) + 100
    
    station_info = n.createVariable('station_info', 'i4', ('station',))
    station_info.long_name = "some kind of station info"
    station_info[...] = [-10, -9, -8]
    
    stationIndex = n.createVariable('stationIndex', 'i4', ('profile',))
    stationIndex.long_name = "which station this profile is for"
    stationIndex.instance_dimension= "station"
    stationIndex[...] = [2, 1, 0, 2, 1, 1, 0, 2, 
                         1, 0, 1, 2, 2, 1, 1,                     
                         2, 1, 0, 2, 1, 1, 0, 2, 
                         1, 0, 1, 2, 2, 1, 1,
                         2, 1, 0, 2, 1, 1, 0, 2, 
                         1, 0, 1, 2, 2, 1, 1,
                         2, 1, 0, 2, 1, 1, 0, 2, 
                         1, 0, 1, 2, 2]
    # station N has list(stationIndex[...]).count(N) profiles
    
    row_size = n.createVariable('row_size', 'i4', ('profile',))
    row_size.long_name = "number of observations for this profile"
    row_size.sample_dimension = "obs"
    row_size[...] = [1, 4, 1, 3, 2, 2, 3, 3, 1, 2, 2, 3, 2, 2, 2, 2, 1, 2, 1, 3, 3, 2, 1,
                     3, 1, 3, 2, 3, 1, 3, 3, 2, 2, 2, 1, 1, 1, 3, 1, 1, 2, 1, 1, 3, 3, 2,
                     2, 2, 2, 1, 2, 3, 3, 3, 2, 3, 1, 1] # sum = 118
    
    time = n.createVariable('time', 'f8', ('profile',))
    time.standard_name = "time";
    time.long_name = "time"
    time.units = "days since 1970-01-01 00:00:00"
    t0 = [3, 0, -3]
    ssi = [0, 0, 0]
    for i, si in enumerate(stationIndex[...]):
        time[i] = t0[si] + ssi[si]
        ssi[si] += 1
        
    z = n.createVariable('z', 'f8', ('obs',))
    z.standard_name = "altitude"
    z.long_name = "height above mean sea level"
    z.units = "km"
    z.axis = "Z"  
    z.positive = "up"
        
#        z0 = [1, 0, 3]
#        i = 0
#        for s, r in zip(stationIndex[...], row_size[...]):
#            z[i:i+r] = z0[s] + numpy.sort(numpy.random.uniform(0, numpy.random.uniform(1, 2), r))
#            i += r

    data =  [3.51977705293769, 0.521185292100177, 0.575154265863394, 
             1.08495843717095, 1.37710968624395, 2.07123455611723, 3.47064474274781, 
             3.88569849023813, 4.81069254279537, 0.264339600625496, 0.915704970094182, 
             0.0701532210336895, 0.395517651420933, 1.00657582854276, 
             1.17721374303641, 1.82189345615046, 3.52424307197668, 3.93200473199559, 
             3.95715099603671, 1.57047493027102, 1.09938982652955, 1.17768722826975, 
             0.251803399458277, 1.59673486865804, 4.02868944763605, 4.03749228832264, 
             4.79858281590985, 3.00019933315412, 3.65124061660449, 0.458463542157766, 
             0.978678197083262, 0.0561560792556281, 0.31182013232255, 
             3.33350065357286, 4.33143904011861, 0.377894196412131, 1.63020681064712, 
             2.00097025264771, 3.76948048424458, 0.572927165845568, 1.29408313557905, 
             1.81296270533192, 0.387142669131077, 0.693459187515738, 1.69261930636298, 
             1.38258797228361, 1.82590759889566, 3.34993297710761, 0.725250730922501, 
             1.38221693486728, 1.59828555215646, 1.59281225554253, 0.452340646918555, 
             0.976663373825433, 1.12640496317618, 3.19366847375422, 3.37209133117904, 
             3.40665008236976, 3.53525896684001, 4.10444186715724, 0.14920937817654, 
             0.0907197953552753, 0.42527916794473, 0.618685137936187, 
             3.01900591447357, 3.37205542289986, 3.86957342976163, 0.17175098751914, 
             0.990040375014957, 1.57011428605984, 2.12140567043994, 3.24374743730506, 
             4.24042441581785, 0.929509749153725, 0.0711997786817564, 
             2.25090028461898, 3.31520955860746, 3.49482624434274, 3.96812568493549, 
             1.5681807261767, 1.79993011515465, 0.068325990211909, 0.124469638352167, 
             3.31990436971169, 3.84766748039389, 0.451973490541035, 1.24303219956085, 
             1.30478004656262, 0.351892459787624, 0.683685812990457, 
             0.788883736575568, 3.73033428872491, 3.99479807507392, 0.811582011950481, 
             1.2241242448019, 1.25563109687369, 2.16603674712822, 3.00010622131408, 
             3.90637137662453, 0.589586644805982, 0.104656387266266, 
             0.961185900148304, 1.05120351477824, 1.29460917520233, 2.10139985693684, 
             3.64252693587415, 3.91197236350995, 4.56466622863717, 0.556476687600461, 
             0.783717448678148, 0.910917550635007, 1.59750076220451, 1.97101264162631, 
             0.714693043642084, 0.904381625638779, 1.03767817888021, 4.10124675852254, 
             3.1059214185543]        
    data = numpy.around(data, 2)
    z[...] = data
    
    humidity = n.createVariable('humidity', 'f8', ('obs',), fill_value=-999.9)
    humidity.standard_name = "specific_humidity"
    humidity.coordinates = "time lat lon alt z station_name station_info profile"
    
    data *= 10
    data = numpy.around(data, 2)
    humidity[...] = data
    
    temp = n.createVariable('temp', 'f8', ('obs',), fill_value=-999.9)
    temp.standard_name = "air_temperature"
    temp.units = "Celsius"
    temp.coordinates = "time lat lon alt z station_name station_info profile"
    
    data += 2731.5
    data = numpy.around(data, 2)
    temp[...] = data
    
    n.close()
    
    return filename
#--- End: def

contiguous_file = _make_contiguous_file('DSG_timeSeries_contiguous.nc')
indexed_file    = _make_indexed_file('DSG_timeSeries_indexed.nc')
indexed_contiguous_file = _make_indexed_contiguous_file('DSG_timeSeriesProfile_indexed_contiguous.nc')

class DSGTest(unittest.TestCase):
    def setUp(self):
        self.contiguous         = contiguous_file
        self.indexed            = indexed_file
        self.indexed_contiguous = indexed_contiguous_file

        (fd, self.tempfilename) = tempfile.mkstemp(suffix='.nc', prefix='cfdm_', dir='.')
        os.close(fd)
        
        a = numpy.ma.masked_all((4, 9), dtype=float)
        a[0, 0:3] = [0.0, 1.0, 2.0]
        a[1, 0:7] = [1.0, 11.0, 21.0, 31.0, 41.0, 51.0, 61.0]
        a[2, 0:5] = [2.0, 102.0, 202.0, 302.0, 402.0]
        a[3, 0:9] = [3.0, 1003.0, 2003.0, 3003.0, 4003.0, 5003.0, 6003.0, 7003.0, 8003.0]
        self.a = a

        b = numpy.array([[[20.7, -99, -99, -99],
                          [10.1, 11.8, 18.2, -99],
                          [11.0, 11.8, -99, -99],
                          [16.3, 20.0, -99, -99],
                          [13.8, 18.3, -99, -99],
                          [15.9 , -99, -99, -99],
                          [15.7, 21.2, -99, -99],
                          [22.5, -99, -99, -99],
                          [18.0, -99, -99, -99],
                          [12.6, 21.7, -99, -99],
                          [10.5, 12.9, 21.0, -99],
                          [16.0, 19.7, -99, -99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99]],
                         
                         [[5.2, 5.8, 10.8, 13.8],
                          [2.6, 9.2, -99, -99],
                          [0.7, 4.0, -99, -99],
                          [15.7, -99, -99, -99],
                          [2.5, 16.0, -99, -99],
                          [4.6, 9.8, -99, -99],
                          [0.6, 3.1, -99, -99],
                          [3.8, -99, -99, -99],
                          [5.7, 12.9, 18.1, -99],
                          [3.9, 6.9, 16.9, -99],
                          [7.3, 13.8, 16.0, -99],
                          [4.5, 9.8, 11.3, -99],
                          [1.5, -99, -99, -99],
                          [0.9, 4.3, 6.2, -99],
                          [1.7, 9.9, -99, -99],
                          [9.3, -99, -99, -99],
                          [0.7, -99, -99, -99],
                          [15.7, -99, -99, -99],
                          [0.7, 1.2, -99, -99],
                          [4.5, 12.4, 13.0, -99],
                          [3.5, 6.8, 7.9, -99],
                          [8.1, 12.2, -99, -99],
                          [5.9, -99, -99, -99],
                          [1.0, 9.6, -99, -99],
                          [5.6, 7.8, 9.1, -99],
                          [7.1, 9.0, 10.4, -99]],
                         
                         [[35.2, -99, -99, -99],
                          [34.7, 38.9, 48.1, -99],
                          [35.2, 39.3, 39.6, -99],
                          [40.3, 40.4, 48.0, -99],
                          [30.0, 36.5, -99, -99],
                          [33.3, 43.3, -99, -99],
                          [37.7, -99, -99, -99],
                          [33.5, -99, -99, -99],
                          [31.9, 33.7, -99, -99],
                          [34.1, 35.4, 41.0, -99],
                          [30.2, 33.7, 38.7, -99],
                          [32.4, 42.4, -99, -99],
                          [33.2, 34.9, 39.7, -99],
                          [33.2, -99, -99, -99],
                          [38.5, -99, -99, -99],
                          [37.3, 39.9, -99, -99],
                          [30.0, 39.1, -99, -99],
                          [36.4, 39.1, 45.6, -99],
                          [41.0, -99, -99, -99],
                          [31.1, -99, -99, -99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99],
                          [-99,-99,-99,-99]]])
        
        b = numpy.ma.where(b==-99, numpy.ma.masked, b)
        self.b = b
        
        self.test_only = []
    #--- End: def
 
    def tearDown(self):
        os.remove(self.tempfilename)
    #--- End: def
    
    def test_DSG_contiguous(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.contiguous, verbose=False)

        self.assertTrue(len(f) == 2)

        # Select the specific humidity field
        q = [g for g in f
             if g.get_property('standard_name') == 'specific_humidity'][0]

        self.assertTrue(cfdm.functions._numpy_allclose(q.get_array(), self.a))

#        print ('\nf\n')
#        for x in f:
#            print(x)
            
        cfdm.write(f, self.tempfilename, verbose=False)
        g = cfdm.read(self.tempfilename)
        
#        print ('\ng\n')
#        for x in g:
#            print(x)

        self.assertTrue(len(g) == len(f))

        for i in range(len(f)):
            self.assertTrue(g[i].equals(f[i], traceback=True))

        # ------------------------------------------------------------
        # Test creation
        # ------------------------------------------------------------
        # Define the ragged array values
        ragged_array = numpy.array([280, 282.5, 281, 279, 278, 279.5],
                                   dtype='float32')
        
        # Define the count array values
        count_array = [2, 4]
        
        # Create the count variable
        count_variable = cfdm.Count(data=cfdm.Data(count_array))
        count_variable.set_property('long_name', 'number of obs for this timeseries')
        
        # Create the contiguous ragged array object
        array = cfdm.RaggedContiguousArray(
                         compressed_array=cfdm.NumpyArray(ragged_array),
                         shape=(2, 4), size=8, ndim=2,
                         count_variable=count_variable)
        
        # Create the field construct with the domain axes and the ragged
        # array
        tas = cfdm.Field()
        tas.properties({'standard_name': 'air_temperature',
                        'units': 'K',
    		        'featureType': 'timeSeries'})
        
        # Create the domain axis constructs for the uncompressed array
        X = tas.set_construct(cfdm.DomainAxis(4))
        Y = tas.set_construct(cfdm.DomainAxis(2))
        
        # Set the data for the field
        tas.set_data(cfdm.Data(array), axes=[Y, X])
        
        cfdm.write(tas, self.tempfilename)
    #--- End: def   
        
        
    #--- End: def        

    def test_DSG_indexed(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.indexed)

        self.assertTrue(len(f) == 2)

        # Select the specific humidity field
        q = [g for g in f
             if g.get_property('standard_name') == 'specific_humidity'][0]

        self.assertTrue(cfdm.functions._numpy_allclose(q.get_array(), self.a))
        
#        print ('\nf\n')
#        for x in f:
#            print(x)
            
        cfdm.write(f, self.tempfilename, verbose=False)
        g = cfdm.read(self.tempfilename)
        
#        print ('\ng\n')
#        for x in g:
#            print(x)

        self.assertTrue(len(g) == len(f))

        for i in range(len(f)):
            self.assertTrue(g[i].equals(f[i], traceback=True))
    #--- End: def        

    def test_DSG_indexed_contiguous(self):  
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.read(self.indexed_contiguous, verbose=False)

        self.assertTrue(len(f) == 2)

        # Select the specific humidity field
        q = [g for g in f
             if g.get_property('standard_name') == 'specific_humidity'][0]

        q = q.get_array()
        
        message= repr(q-self.b) +'\n'+repr(q[2,0])+'\n'+repr(self.b[2, 0])
        self.assertTrue(cfdm.functions._numpy_allclose(q, self.b),
                        message)        

#        print ('\nf\n')
#        for x in f:
#            print(x)
            
        cfdm.write(f, self.tempfilename, verbose=False)
        g = cfdm.read(self.tempfilename, verbose=False)
        
#        print ('\ng\n')
#        for x in g:
#            print(x)

        self.assertTrue(len(g) == len(f))

        for i in range(len(f)):
            self.assertTrue(g[i].equals(f[i], traceback=True))
    #--- End: def        

    def test_DSG_create_contiguous(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return
        
        # Define the ragged array values
        ragged_array = numpy.array([1, 3, 4, 3, 6], dtype='float32')
        # Define the count array values
        count_array = [2, 3]
        
        # Initialise the count variable
        count_variable = cfdm.Count(data=cfdm.Data(count_array))
        count_variable.set_property('long_name', 'number of obs for this timeseries')
        
        # Initialise the contiguous ragged array object
        array = cfdm.RaggedContiguousArray(
            compressed_array=cfdm.NumpyArray(ragged_array),
            shape=(2, 3), size=6, ndim=2,
            count_variable=count_variable)
        
        # Initialize the auxiliary coordinate construct with the ragged
        # array and set some properties
        z = cfdm.AuxiliaryCoordinate(
            data=cfdm.Data(array),
            properties={'standard_name': 'height',
                        'units': 'km',
                        'positive': 'up'})
        
        self.assertTrue((z.get_array() == numpy.ma.masked_array(
            data=[[1.0, 3.0, 99],
                  [4.0, 3.0, 6.0]],
            mask=[[False, False,  True],
                  [False, False, False]],
            fill_value=1e+20,
            dtype='float32')).all())
        
        self.assertTrue(z.data.get_compression_type() == 'ragged contiguous')
        
        self.assertTrue((z.data.get_compressed_array() == numpy.array(
            [1., 3., 4., 3., 6.], dtype='float32')).all())
        
        self.assertTrue((z.data.get_count_variable().get_array() == numpy.array(
            [2, 3])).all())
    #--- End: def

#--- End: class


if __name__ == '__main__':
    print('Run date:', datetime.datetime.utcnow())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)
    



from builtins import (str, super)

import numpy

from .. import abstract


class Data(abstract.Container):
    '''An orthogonal multidimensional array with masked values and units.

.. versionadded:: 1.7

    '''
    def __init__(self, array=None, units=None, calendar=None,
                 fill_value=None, source=None, copy=True,
                 _use_array=True):
        '''**Initialization**

:Parameters:

    array: subclass of `Array`
        The array of values. Ignored if the *source* parameter is set.

    units: `str`, optional
        The physical units of the data. Ignored if the *source*
        parameter is set.

        *Example:*
          ``units='km hr-1'``

        *Example:*
          ``units='days since 2018-12-01'``

        The units may also be set after initialisation with the
        `set_units` method.

    calendar: `str`, optional
        The calendar for reference time units. Ignored if the *source*
        parameter is set.

        *Example:*
          ``calendar='360_day'``
        
        The calendar may also be set after initialisation with the
        `set_calendar` method.

    fill_value: optional 
        The fill value of the data. By default, or if None, the numpy
        fill value appropriate to the array's data type will be used.
        TODO. Ignored if the *source* parameter is set.

        *Example:*
          ``fill_value=-999.``
                
        The fill value may also be set after initialisation with the
        `set_fill_value` method.

    source: *optional*
        Initialize the data, units, calendar and fill value from those
        of *source*.
        
    source: optional
        Initialize the array, units, calendar and fill value from
        those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''
        super().__init__()
        
        if source is not None:
            try:                
                array = source._get_Array(None)
            except AttributeError:
                array = None

            try:
                units = source.get_units(None)
            except AttributeError:
                units = None
                
            try:
                calendar = source.get_calendar(None)
            except AttributeError:
                calendar = None

            try:
                fill_value = source.get_fill_value(None)
            except AttributeError:
                fill_value = None
        #--- End: if

        self.set_units(units)
        self.set_calendar(calendar)   
        self.set_fill_value(fill_value)

        if _use_array:
            self._set_Array(array, copy=copy)
    #--- End: def

#    def __array__(self, *dtype):
#        '''The numpy array interface.
#
#:Returns: 
#
#    out: `numpy.ndarray`
#        An independent numpy array of the data.
#
#        '''
#        array = self.get_array()
#        if not dtype:
#            return array
#        else:
#            return array.astype(dtype[0], copy=False)
#    #--- End: def

#    def __deepcopy__(self, memo):
#        '''x.__deepcopy__() -> Deep copy of data.
#
#Used if copy.deepcopy is called on the object.
#
#        ''' 
#        return self.copy()
#    #--- End: def

#    def __repr__(self):
#        '''x.__repr__() <==> repr(x)
#
#        '''
#        try:        
#            shape = self.shape
#        except AttributeError:
#            shape = ''
#        else:
#            shape = str(shape)
#            shape = shape.replace(',)', ')')
#            
#        return '<{0}{1}: {2}>'.format(self.__class__.__name__, shape, str(self))
#    #--- End: def
#
#    def __str__(self):
#        '''x.__str__() <==> str(x)
#
#        '''
#        return str(self._get_Array(None))
#    #--- End: def
    
    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def dtype(self):
        '''Data-type of the data elements.

:Examples:

>>> d.dtype
dtype('float64')
>>> type(d.dtype)
<type 'numpy.dtype'>

        '''
        return self._get_Array().dtype        
    #--- End: def

    @property
    def ndim(self):
        '''Number of data dimensions.

:Examples:

>>> d.shape
(73, 96)
>>> d.ndim
2
>>> d.size
7008

>>> d.shape
(1, 1, 1)
>>> d.ndim
3
>>> d.size
1

>>> d.shape
()
>>> d.ndim
0
>>> d.size
1

        '''
        return self._get_Array().ndim
    #--- End: def

    @property
    def shape(self):
        '''Tuple of data dimension sizes.

:Examples:

>>> d.shape
(73, 96)
>>> d.ndim
2
>>> d.size
7008

>>> d.shape
(1, 1, 1)
>>> d.ndim
3
>>> d.size
1

>>> d.shape
()
>>> d.ndim
0
>>> d.size
1

        '''
        return self._get_Array().shape
    #--- End: def

    @property
    def size(self):
        '''Number of elements in the data.

:Examples:

>>> d.shape
(73, 96)
>>> d.size
7008
>>> d.ndim
2

>>> d.shape
(1, 1, 1)
>>> d.ndim
3
>>> d.size
1

>>> d.shape
()
>>> d.ndim
0
>>> d.size
1

        '''
        return self._get_Array().size
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self, array=True):
        '''Return a deep copy of the data.

``d.copy()`` is equivalent to ``copy.deepcopy(d)``.

Copy-on-write is employed, so care must be taken when modifying any
attribute.

:Parameters:

    array: `bool`, optional
        If False then do not copy the array. By default the array is
        copied.

:Returns:

    out: 
        The deep copy.

:Examples:

>>> e = d.copy()
>>> e = d.copy(array=False)

        '''
        return type(self)(source=self, copy=True, _use_array=array)
    #--- End: def

    def del_calendar(self):
        '''Delete the calendar.

.. seealso:: `get_calendar`, `set_calendar`

:Examples 1:

>>> u = d.del_calendar()

:Returns:

    out:
        The value of the deleted calendar, or `None` if calendar was
        not set.

:Examples 2:

>>> d.set_calendar('proleptic_gregorian')
>>> d.get_calendar
'proleptic_gregorian'
>>> d.del_calendar()
>>> d.get_calendar()
AttributeError: Can't get non-existent calendar
>>> print(d.get_calendar(None))
None

        '''
        return self._del_component('calendar')
#        value = self._calendar
#        self._calendar = None
#        return value
    #--- End: def

    def _del_Array(self):
        '''Delete the data.

:Examples 1:

>>> d.del_data()

:Returns:

    out:

:Examples 2:

>>> old = d.del_data()

        '''
        return self._del_component('data')
    #--- End: def
    
    def del_fill_value(self):
        '''Delete the fill value.

.. seealso:: `get_fill_value`, `set_fill_value`

:Examples 1:

>>> f = d.del_fill_value()

:Returns:

    out:
        The value of the deleted fill value, or `None` if fill value
        was not set.

:Examples 2:

>>> f.set_fill_value(-9999)
>>> f.get_fill_value()
-9999
>>> print(f.del_fill_value())
-9999
>>> f.get_fill_value()
AttributeError: Can't get non-existent fill value
>>> f.get_fill_value(10**10)
10000000000
>>> print(f.get_fill_value(None))
None
>>> f.set_fill_value(None)
>>> print(f.get_fill_value())
None

        '''
        return self._del_component('fill_value')
#        value = self._fill_value
#        self._fill_value = None        
#        return value
    #--- End: def

    def del_units(self):
        '''Delete the units.

.. seealso:: `get_units`, `set_units`

:Examples 1:

>>> u = d.del_units()

:Returns:

    out:
        The value of the deleted units, or `None` if units was not
        set.

:Examples 2:

>>> d.set_units('metres')
>>> d.get_units()
'metres'
>>> d.del_units()
>>> d.get_units()
AttributeError: Can't get non-existent units
>>> print(d.get_units(None))
None

        '''
        return self._del_component('units')
#        value = self._units
#        self._units = None        
#        return value
    #--- End: def

    def get_array(self):
        '''Return an independent numpy array containing the data.

If a fill value has been set (see `set_fill_value`) then it will be
used, otherwise the default numpy fill value appropriate to the data
type will be used.

:Returns:

    out: `numpy.ndarray`
        An independent numpy array of the data.

:Examples:

>>> d = Data([1, 2, 3.0], 'km')
>>> n = d.get_array()
>>> isinstance(n, numpy.ndarray)
True
>>> print(n)
[ 1.,   2.,   3.]
>>> n[0] = 88
>>> print(repr(d))
<Data: [1.0, 2.0, 3.0] km>

        '''
        array = self._get_Array().get_array()

        # Set the numpy array fill value
        if numpy.ma.isMA(array):
            array.set_fill_value(self.get_fill_value(None))

        return array
    #--- End: def

    def get_calendar(self, *default):
        '''Return the calendar.

.. seealso:: `del_calendar`, `set_calendar`

:Examples 1:

>>> u = d.get_calendar()

:Parameters:

    default: optional
        Return *default* if calendar has not been set.

:Returns:

    out:
        The calendar. If calendar has not been set then return the
        value of *default* parameter, if provided.

:Examples 2:

>>> d.set_calendar('julian')
>>> d.get_calendar
'metres'
>>> d.del_calendar()
>>> d.get_calendar()
AttributeError: Can't get non-existent calendar
>>> print(d.get_calendar(None))
None

        '''
        value = self._get_component('calendar', *default)
#        value = self._calendar
        if value is None:
            if default:
                return default[0]

            raise AttributeError("{!r} has no calendar".format(
                self.__class__.__name__))
        
        return value
    #--- End: def

    def _get_Array(self, *default):
        '''Return the data.

:Examples 1:

>>> a = d.get_data()

:Parameters:

    default: *optional*
        If there is no data then *default* if returned if set.

:Returns:

    out:
        The data. If the data has not been set then *default* is
        returned, if set.

:Examples 2:

>>> a = d.get_data(None)

        '''
        array = self._get_component('array', *default)

        if array is None:
            if default:
                return default[0]     

            raise AttributeError("{!r} has no array".format(
                self.__class__.__name__))
        
        return array   
    #--- End: def

    def get_fill_value(self, *default):
        '''Return the missing data value.

.. seealso:: `del_fill_value`, `set_fill_vlaue`

:Examples 1:

>>> f = d.get_fill_value()

:Parameters:

    default: optional
        Return *default* if fill value has not been set.

:Returns:

    out:
        The fill value. If fill value has not been set then return the
        value of *default* parameter, if provided.

:Examples 2:

>>> f.set_fill_value(-9999)
>>> f.get_fill_value()
-9999
>>> print(f.del_fill_value())
-9999
>>> f.get_fill_value()
AttributeError: Can't get non-existent fill value
>>> f.get_fill_value(10**10)
10000000000
>>> print(f.get_fill_value(None))
None
>>> f.set_fill_value(None)
>>> print(f.get_fill_value())
None

        '''
        value = self._get_component('fill_value', *default)
#        value = self._fill_value
        if value is None:
            if default:
                return default[0]

            raise AttributeError("{!r} has no fill value".format(
                self.__class__.__name__))
        
        return value
    #--- End: def

    def get_units(self, *default):
        '''Return the units.

.. seealso:: `del_units`, `set_units`

:Examples 1:

>>> u = d.get_units()

:Parameters:

    default: optional
        Return *default* if units has not been set.

:Returns:

    out:
        The units. If units has not been set then return the value of
        *default* parameter, if provided.

:Examples 2:

>>> d.set_units('metres')
>>> d.get_units()
'metres'
>>> d.del_units()
>>> d.get_units()
AttributeError: Can't get non-existent units
>>> print(d.get_units(None))
None

        '''
        value = self._get_component('units', *default)
#        value = self._units
        if value is None:
            if default:
                return default[0]

            raise AttributeError("{!r} has no units".format(
                self.__class__.__name__))
        
        return value
    #--- End: def

    def set_calendar(self, calendar):
        '''Set the calendar.

.. seealso:: `del_calendar`, `get_calendar`

:Examples 1:

>>> u = d.set_calendar('365_day')

:Parameters:

    value: `str`
        The new calendar.

:Returns:

    `None`

:Examples 2:

>>> d.set_calendar('none')
>>> d.get_calendar
'none'
>>> d.del_calendar()
>>> d.get_calendar()
AttributeError: Can't get non-existent calendar
>>> print(d.get_calendar(None))
None

        '''
        return self._set_component('calendar', calendar, copy=False)
#        self._calendar = calendar
    #--- End: def

    def _set_Array(self, array, copy=True):
        '''Set the array.

:Parameters:

    array: subclass of `Array`
        The array to be inserted.

:Returns:

    `None`

:Examples:

>>> d._set_Array(a)

        '''
        if copy:
            array = array.copy()
            
        self._set_component('array', array, copy=False)
    #--- End: def

    def set_fill_value(self, value):
        '''Set the missing data value.

.. seealso:: `del_fill_value`, `get_fill_vlaue`

:Examples 1:

>>> f = d.set_fill_value(-256)

:Parameters:

    value: scalar
        The new fill value.

:Returns:

    `None`

:Examples 2:

>>> f.set_fill_value(-9999)
>>> f.get_fill_value()
-9999
>>> print(f.del_fill_value())
-9999
>>> f.get_fill_value()
AttributeError: Can't get non-existent fill value
>>> f.get_fill_value(10**10)
10000000000
>>> print(f.get_fill_value(None))
None
>>> f.set_fill_value(None)
>>> print(f.get_fill_value())
None

        '''
#        self._fill_value = value
        self._set_component('fill_value', value, copy=False)
    #--- End: def

    def set_units(self, value):
        '''Set the units.

.. seealso:: `del_units`, `get_units`

:Examples 1:

>>> u = d.set_units('kg m-2')

:Parameters:

    value: `str`
        The new units.

:Returns:

    `None`

:Examples 2:

>>> d.set_units('watt')
>>> d.get_units()
'watt'
>>> d.del_units()
>>> d.get_units()
AttributeError: Can't get non-existent units
>>> print(d.get_units(None))
None

        '''
        self._set_component('units', value, copy=False)
#        self._units = value
    #--- End: def

#--- End: class




from __future__ import print_function
from builtins import range
import datetime
import inspect
import itertools
from operator import mul
import os
import sys
import time
import unittest


import numpy

import cfdm

class DataTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    test_only = []
#    test_only = ['NOTHING!!!!!']
#    test_only = ['test_Data__asdatetime__asreftime__isdatetime']
#    test_only = ['test_Data___setitem__']
#    test_only = ['test_Data_ceil', 'test_Data_floor', 'test_Data_trunc', 'test_Data_rint']
#    test_only = ['test_Data_array', 'test_Data_dtarray']
#    test_only = ['test_dumpd_loadd']
#    test_only = ['test_Data_BINARY_AND_UNARY_OPERATORS']

    def test_Data___getitem__(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return
        
    def test_Data___setitem__(self):        
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        a = numpy.ma.arange(3000).reshape(50, 60)

        d = cfdm.Data(a.filled(), units='m')
        
        for n, (j, i) in enumerate(((34, 23), (0, 0), (-1, -1),
                                    (slice(40, 50), slice(58, 60)),
                                    (Ellipsis, slice(None)),
                                    (slice(None), Ellipsis),
                                  )):
            n = -n-1
            for dvalue, avalue in ((n, n), (cfdm.masked, numpy.ma.masked), (n, n)):
                message = "cfdm.Data[{}, {}]={}={} failed".format(j, i, dvalue, avalue)
                d[j, i] = dvalue
                a[j, i] = avalue
                x = d.get_array()
                self.assertTrue((x == a).all() in (True, numpy.ma.masked), message)
                m = numpy.ma.getmaskarray(x)
                self.assertTrue((m == numpy.ma.getmaskarray(a)).all(), 
                                'd.mask.array='+repr(m)+'\nnumpy.ma.getmaskarray(a)='+repr(numpy.ma.getmaskarray(a)))
        #--- End: for
    
        a = numpy.ma.arange(3000).reshape(50, 60)
        
        d = cfdm.Data(a.filled(), 'm')
        
        (j, i) = (slice(0, 2), slice(0, 3))
        array = numpy.array([[1, 2, 6],[3, 4, 5]])*-1
        
        for dvalue in (array,
                       numpy.ma.masked_where(array < -2, array),
                       array):
            message = 'cfdm.Data[%s, %s]=%s failed' % (j, i, dvalue)
            d[j, i] = dvalue
            a[j, i] = dvalue
            x = d.get_array()
            self.assertTrue((x == a).all() in (True, numpy.ma.masked), message)
            m = numpy.ma.getmaskarray(x)
            self.assertTrue((m == numpy.ma.getmaskarray(a)).all(), message)
        #--- End: for
        
        # Scalar numeric array
        d = cfdm.Data(9, units='km')
        d[...] = cfdm.masked
        a = d.get_array()
        self.assertTrue(a.shape == ())
        self.assertTrue(a[()] is numpy.ma.masked)
    #--- End: def

    def test_Data_get_array(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Scalar numeric array
        d = cfdm.Data(9, units='km')
        a = d.get_array()
        self.assertTrue(a.shape == ())
        self.assertTrue(a == numpy.array(9))
        d[...] = cfdm.masked
        a = d.get_array()
        self.assertTrue(a.shape == ())
        self.assertTrue(a[()] is numpy.ma.masked)

        # Non-scalar numeric array
        b = numpy.arange(10*15*19).reshape(10, 1, 15, 19)
        d = cfdm.Data(b, 'km')
        a = d.get_array()
        a[0,0,0,0] = -999
        a2 = d.get_array()
        self.assertTrue(a2[0,0,0,0] == 0)
        self.assertTrue(a2.shape == b.shape)
        self.assertTrue((a2 == b).all())
        self.assertFalse((a2 == a).all())
    #--- End: def

    def test_Data_get_dtarray(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        d = cfdm.Data(11292.5, units='days since 1970-1-1')
        dt = d.get_dtarray()[()]
        self.assertTrue(dt == datetime.datetime(2000, 12, 1, 12, 0))
    #--- End: def

    def test_Data_transpose(self):        
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        a = numpy.arange(10*15*19).reshape(10, 1, 15, 19)
        d = cfdm.Data(a.copy())
        
        for indices in (list(range(a.ndim)), list(range(-a.ndim, 0))):
            for axes in itertools.permutations(indices):
                a = numpy.transpose(a, axes)
                d = d.transpose(axes)
                message = 'cfdm.Data.transpose({}) failed: d.shape={}, a.shape={}'.format(
                    axes, d.shape, a.shape)
                self.assertTrue(d.shape == a.shape, message)
                self.assertTrue((d.get_array() == a).all(), message)
            #--- End: for
        #--- End: for
    #--- End: def
        
    def test_Data_unique(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return
            
        d = cfdm.Data([[4, 2, 1], [1, 2, 3]], units='metre')
        u = d.unique()
        self.assertTrue(u.shape == (4,))
        self.assertTrue((u.get_array() == cfdm.Data([1, 2, 3, 4], 'metre').get_array()).all())

        d[1, -1] = cfdm.masked
        u = d.unique()
        self.assertTrue(u.shape == (3,))        
        self.assertTrue((u.get_array() == cfdm.Data([1, 2, 4], 'metre').get_array()).all())
    #--- End: def

#--- End: class


if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cfdm.environment()
    print('')
    unittest.main(verbosity=2)



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import zip
from builtins import str
import os
import sys

from collections import Iterable

from itertools   import zip_longest, product
from platform    import system, platform, python_version
from urllib.parse    import urlparse as urlparse_urlparse
from urllib.parse    import urljoin  as urlparse_urljoin

import netCDF4
import numpy
import future

from . import __version__, __cf_version__, __file__

from .constants import CONSTANTS

def _numpy_allclose(a, b, rtol=None, atol=None):
    '''

Returns True if two broadcastable arrays have equal values to within
numerical tolerance, False otherwise.

The tolerance values are positive, typically very small numbers. The
relative difference (``rtol * abs(b)``) and the absolute difference
``atol`` are added together to compare against the absolute difference
between ``a`` and ``b``.

:Parameters:

    a, b : array_like
        Input arrays to compare.

    atol : float, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol : float, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

:Returns:

    out: `bool`
        Returns True if the arrays are equal, otherwise False.

**Examples**

>>> cf._numpy_allclose([1, 2], [1, 2])
True
>>> cf._numpy_allclose(numpy.array([1, 2]), numpy.array([1, 2]))
True
>>> cf._numpy_allclose([1, 2], [1, 2, 3])
False
>>> cf._numpy_allclose([1, 2], [1, 4])
False

>>> a = numpy.ma.array([1])
>>> b = numpy.ma.array([2])
>>> a[0] = numpy.ma.masked
>>> b[0] = numpy.ma.masked
>>> cf._numpy_allclose(a, b)
True

'''

    # THIS IS WHERE SOME NUMPY FUTURE WARNINGS ARE COMING FROM
    
    a_is_masked = numpy.ma.isMA(a)
    b_is_masked = numpy.ma.isMA(b)
#    print 'a_is_masked   , b_is_masked ', a_is_masked   , b_is_masked    
    if not (a_is_masked or b_is_masked):
        try:            
            return numpy.allclose(a, b, rtol=rtol, atol=atol)
        except (IndexError, NotImplementedError, TypeError):
            return numpy.all(a == b)
    else:
        if a_is_masked and b_is_masked:
            # a and b are both masked arrays
            if (a.mask != b.mask).any():
                return False
        elif not ((a_is_masked and not a.mask.any()) or
                  (b_is_masked and not b.mask.any())):
            return False

        try:
            return numpy.ma.allclose(a, b, rtol=rtol, atol=atol)
        except (IndexError, NotImplementedError, TypeError):
            out = numpy.ma.all(a == b)
            if out is numpy.ma.masked:
                return True
            else:
                return out
#--- End: def

def _numpy_isclose(a, b, rtol=None, atol=None):
    '''Returns a boolean array where two broadcastable arrays are
element-wise equal within a tolerance.

The tolerance values are positive, typically very small numbers. The
relative difference (``rtol * abs(b)``) and the absolute difference
``atol`` are added together to compare against the absolute difference
between ``a`` and ``b``.

:Parameters:

    a, b: array_like
        Input arrays to compare.

    atol: `float`, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol: `float`, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

:Returns:

    out: `numpy.ndarray`

**Examples**
    '''      
    try:
        return numpy.isclose(a, b, rtol=rtol, atol=atol)
    except (IndexError, NotImplementedError, TypeError):
        return a == b
#--- End: def

#def parse_indices(shape, indices):
#    '''
#
#:Parameters:
#
#    shape: sequence of `ints`
#
#    indices: `tuple` (not a `list`!)
#
#:Returns:
#
#    out: `list`
#
#**Examples**
#
#'''
#    parsed_indices = []
#    roll           = {}
#    flip           = []
#    compressed_indices = []
#
#    if not isinstance(indices, tuple):
#        indices = (indices,)
#
#    # Initialize the list of parsed indices as the input indices with any
#    # Ellipsis objects expanded
#    length = len(indices)
#    n = len(shape)
#    ndim = n
#    for index in indices:
#        if index is Ellipsis:
#            m = n - length + 1
#            parsed_indices.extend([slice(None)] * m)
#            n -= m            
#        else:
#            parsed_indices.append(index)
#            n -= 1
#
#        length -= 1
#    #--- End: for
#    len_parsed_indices = len(parsed_indices)
#
#    if ndim and len_parsed_indices > ndim:
#        raise IndexError("Invalid indices %s for array with shape %s" %
#                         (parsed_indices, shape))
#
#    if len_parsed_indices < ndim:
#        parsed_indices.extend([slice(None)]*(ndim-len_parsed_indices))
#
#    if not ndim and parsed_indices:
#        raise IndexError("Scalar array can only be indexed with () or Ellipsis")
#
#    for i, (index, size) in enumerate(zip(parsed_indices, shape)):
#        is_slice = False
#
#        if isinstance(index, slice):            
#            is_slice = True
#
#        elif isinstance(index, (int, int)):
#            if index < 0: 
#                index += size
#
#            index = slice(index, index+1, 1)
#            is_slice = True
#        else:
#            convert2positve = True
#            if getattr(getattr(index, 'dtype', None), 'kind', None) == 'b':
#                # Convert booleans to non-negative integers. We're
#                # assuming that anything with a dtype attribute also
#                # has a size attribute.
#                if index.size != size:
#                    raise IndexError(
#"Invalid indices {} for array with shape {}".format(parsed_indices, shape))
#                
#                index = numpy.where(index)[0]
#                convert2positve = False
#            #--- End: if
#
#            if not numpy.ndim(index):
#                if index < 0:
#                    index += size
#
#                index = slice(index, index+1, 1)
#                is_slice = True
#            else:
#                len_index = len(index)
#                if len_index == 1:                
#                    index = index[0]
#                    if index < 0:
#                        index += size
#                    
#                    index = slice(index, index+1, 1)
#                    is_slice = True
#                elif len_index:
#                    if convert2positve:
#                        # Convert to non-negative integer numpy array
#                        index = numpy.array(index)
#                        index = numpy.where(index < 0, index+size, index)
#    
#                    steps = index[1:] - index[:-1]
#                    step = steps[0]
#                    if step and not (steps - step).any():
#                        # Replace the numpy array index with a slice
#                        if step > 0:
#                            start, stop = index[0], index[-1]+1
#                        elif step < 0:
#                            start, stop = index[0], index[-1]-1
#                            
#                        if stop < 0:
#                            stop = None
#                                
#                        index = slice(start, stop, step)
#                        is_slice = True
#                else:
#                    raise IndexError(
#                        "Invalid indices {} for array with shape {}".format(
#                            parsed_indices, shape))                
#            #--- End: if
#        #--- End: if
#        
#        parsed_indices[i] = index    
#    #--- End: for
#
#    return parsed_indices
##--- End: def

#def get_subspace(array, indices):
#    '''
#
#:Parameters:
#
#    array : numpy array
#
#    indices : list
#
#Subset the input numpy array with the given indices. Indexing is similar to
#that of a numpy array. The differences to numpy array indexing are:
#
#1. An integer index i takes the i-th element but does not reduce the rank of
#   the output array by one.
#
#2. When more than one dimension's slice is a 1-d boolean array or 1-d sequence
#   of integers then these indices work independently along each dimension
#   (similar to the way vector subscripts work in Fortran).
#
#indices must contain an index for each dimension of the input array.
#'''
#    gg = [i for i, x in enumerate(indices) if not isinstance(x, slice)]
#    len_gg = len(gg)
#
#    if len_gg < 2:
#        # ------------------------------------------------------------
#        # At most one axis has a list-of-integers index so we can do a
#        # normal numpy subspace
#        # ------------------------------------------------------------
#        return array[tuple(indices)]
#
#    else:
#        # ------------------------------------------------------------
#        # At least two axes have list-of-integers indices so we can't
#        # do a normal numpy subspace
#        # ------------------------------------------------------------
#        if numpy.ma.isMA(array):
#            take = numpy.ma.take
#        else:
#            take = numpy.take
#
#        indices = indices[:]
#        for axis in gg:
#            array = take(array, indices[axis], axis=axis)
#            indices[axis] = slice(None)
#        #--- End: for
#
#        if len_gg < len(indices):
#            array = array[tuple(indices)]
#
#        return array
#    #--- End: if
#
##--- End: def

def ATOL(*atol):
    '''The tolerance on absolute differences when testing for numerically
tolerant equality.

Two numbers ``x`` and ``y`` are considered equal if ``abs(x-y) <= atol
+ rtol*abs(y)``, where atol (the tolerance on absolute differences)
and rtol (the tolerance on relative differences) are positive,
typically very small numbers. By default both are set to the system
epsilon (the difference between 1 and the least value greater than 1
that is representable as a float).

.. versionadded:: 1.7

.. seealso:: `cfdm.RTOL`

:Parameters:

    atol: `float`, optional
        The new value of absolute tolerance. The default is to not
        change the current value.

:Returns:

    out: `float`
        The value prior to the change, or the current value if no
        new value was specified.

**Examples**

>>> ATOL()
2.220446049250313e-16
>>> old = ATOL(1e-10)
>>> ATOL(old)
1e-10
>>> ATOL()
2.220446049250313e-16

    '''
    old = CONSTANTS['ATOL']
    if atol:
        CONSTANTS['ATOL'] = atol[0]

    return old
#--- End: def

def RTOL(*rtol):    
    '''The tolerance on relative differences when testing for numerically
tolerant equality.

Two numbers ``x`` and ``y`` are considered equal if ``abs(x-y) <= atol
+ rtol*abs(y)``, where atol (the tolerance on absolute differences)
and rtol (the tolerance on relative differences) are positive,
typically very small numbers. By default both are set to the system
epsilon (the difference between 1 and the least value greater than 1
that is representable as a float).

.. versionadded:: 1.7

.. seealso:: `cfdm.ATOL`

:Parameters:

    rtol: `float`, optional
        The new value of relative tolerance. The default is to not
        change the current value.

:Returns:

    out: `float`
        The value prior to the change, or the current value if no
        new value was specified.

**Examples**

>>> cfdm.RTOL()
2.220446049250313e-16
>>> old = RTOL(1e-10)
>>> RTOL(old)
1e-10
>>> cfdm.RTOL()
2.220446049250313e-16

    '''
    old = CONSTANTS['RTOL']
    if rtol:
        CONSTANTS['RTOL'] = rtol[0]

    return old
#--- End: def

#def equals(x, y, rtol=None, atol=None, **kwargs): #ignore_data_type=False,
##           ignore_fill_value=False, ignore_type=False,
##           traceback=False):
#    '''True if and only if two objects are logically equal.
#
#If the first argument, *x*, has an :meth:`equals` method then it is
#used, and in this case ``equals(x, y)`` is equivalent to
#``x.equals(y)``. Else if the second argument, *y*, has an
#:meth:`equals` method then it is used, and in this case ``equals(x,
#y)`` is equivalent to ``y.equals(x)``.
#
#:Parameters:
#
#    x, y :
#        The objects to compare for equality.
#
#    atol : float, optional
#        The absolute tolerance for all numerical comparisons, By
#        default the value returned by the `ATOL` function is used.
#
#    rtol : float, optional
#        The relative tolerance for all numerical comparisons, By
#        default the value returned by the `RTOL` function is used.
#
#    ignore_fill_value : bool, optional
#        If True then `Data` arrays with different fill values are
#        considered equal. By default they are considered unequal.
#
#    traceback : bool, optional
#        If True then print a traceback highlighting where the two
#        objects differ.
#
#:Returns:
#
#    out: `bool`
#        Whether or not the two objects are equal.
#
#**Examples**
#
#>>> f
#<CF Field: rainfall_rate(latitude(10), longitude(20)) kg m2 s-1>
#>>> cfdm.equals(f, f)
#True
#
#>>> cfdm.equals(1.0, 1.0)
#True
#>>> cfdm.equals(1.0, 33)
#False
#
#>>> cfdm.equals('a', 'a')
#True
#>>> cfdm.equals('a', 'b')
#False
#
#>>> type(x), x.dtype
#(<type 'numpy.ndarray'>, dtype('int64'))
#>>> y = x.copy()
#>>> cfdm.equals(x, y)
#True
#>>> cfdm.equals(x, x+1)
#False
#
#>>> class A(object):
#...     pass
#...
#>>> a = A()
#>>> b = A()
#>>> cfdm.equals(a, a)
#True
#>>> cfdm.equals(a, b)
#False
#
#    '''
#    eq = getattr(x, 'equals', None)
#
#    if callable(eq):
#        # x has a callable equals method
#        return eq(y, rtol=rtol, atol=atol, **kwargs)
#
#    eq = getattr(y, 'equals', None)
#    if callable(eq):
#        # y has a callable equals method
#        return eq(x, rtol=rtol, atol=atol, **kwargs)
# 
#    if isinstance(x, numpy.ndarray) or isinstance(y, numpy.ndarray):
#        if numpy.shape(x) != numpy.shape(y):
#            return False
#
#        if rtol is None:
#            rtol = RTOL()
#        if atol is None:
#            atol = ATOL()
#                
#        return _numpy_allclose(x, y, rtol=rtol, atol=atol)
#
#    else:
#        return x == y
##--- End: def

#def abspath(filename):
#    '''
#
#Return a normalized absolute version of a file name.
#
#If a string containing URL is provided then it is returned unchanged.
#
#:Parameters:
#
#    filename: `str`
#        The name of the file.
#
#:Returns:
#
#    out: `str`
#        The normalized absolutized version of *filename*.
# 
#**Examples**
#
#>>> import os
#>>> os.getcwd()
#'/data/archive'
#>>> abspath('file.nc')
#'/data/archive/file.nc'
#>>> abspath('..//archive///file.nc')
#'/data/archive/file.nc'
#>>> abspath('http://data/archive/file.nc')
#'http://data/archive/file.nc'
#
#'''
#    u = urlparse_urlparse(filename)
#    if u.scheme != '':
#        return filename
#
#    return os.path.abspath(filename)
##--- End: def

def allclose(x, y, rtol=None, atol=None):
    '''

Returns True if two broadcastable arrays have equal values to within
numerical tolerance, False otherwise.

The tolerance values are positive, typically very small numbers. The
relative difference (``rtol * abs(b)``) and the absolute difference
``atol`` are added together to compare against the absolute difference
between ``a`` and ``b``.

:Parameters:

    x, y: array_like
        Input arrays to compare.

    atol: `float`, optional
        The absolute tolerance for all numerical comparisons, By
        default the value returned by the `ATOL` function is used.

    rtol: `float`, optional
        The relative tolerance for all numerical comparisons, By
        default the value returned by the `RTOL` function is used.

:Returns:

    out: `bool`
        Returns True if the arrays are equal, otherwise False.

**Examples**

'''    
    if rtol is None:
        rtol = RTOL()
    if atol is None:
        atol = ATOL()

    allclose = getattr(x, 'allclose', None)
    if callable(allclose):
        # x has a callable allclose method
       return allclose(y, rtol=rtol, atol=atol)

    allclose = getattr(y, 'allclose', None)
    if callable(allclose):
        # y has a callable allclose method
       return allclose(x, rtol=rtol, atol=atol)

    # x nor y has a callable allclose method
    return _numpy_allclose(x, y, rtol=rtol, atol=atol)
#--- End: def

def environment(display=True):
    '''Return the names, versions and paths of all dependencies.

.. versionadded:: 1.7

:Parameters:

    display: `bool`, optional
        If False then return the description of the environment as a
        string. By default the description is printed.

:Returns:

    out:
        If *display* is True then the description of the environment
        is printed and `None` is returned. Otherwise the description
        is returned as a string.

**Examples:**

>>> cfdm.environment()
Platform: Linux-4.15.0-38-generic-x86_64-with-debian-stretch-sid
python: 2.7.15 /home/user/anaconda2/bin/python
HDF5 library: 1.10.1
netcdf library: 4.6.1
netCDF4: 1.4.0 /home/user/anaconda2/lib/python2.7/site-packages/netCDF4/__init__.pyc
numpy: 1.11.3 /home/user/anaconda2/lib/python2.7/site-packages/numpy/__init__.pyc
future: 0.16.0 /home/user/anaconda2/lib/python2.7/site-packages/future/__init__.pyc
cfdm: 1.7.1 /home/user/cfdm/cfdm/__init__.pyc

    '''
    out = []
    out.append('Platform: ' + str(platform()))
    out.append('python: ' + str(python_version() + ' ' + str(sys.executable)))

    out.append('HDF5 library: ' + str(netCDF4. __hdf5libversion__))
    out.append('netcdf library: ' + str(netCDF4.__netcdf4libversion__))

    out.append('netCDF4: ' + str(netCDF4.__version__) + ' ' + str(os.path.abspath(netCDF4.__file__)))
    out.append('numpy: ' + str(numpy.__version__) + ' ' + str(os.path.abspath(numpy.__file__)))
    out.append('future: ' + str(future.__version__) + ' ' + str(os.path.abspath(future.__file__)))

    out.append('cfdm: ' + str(__version__) + ' ' + str(os.path.abspath(__file__)))
    
    out = '\n'.join(out)

    if display:
        print(out)
    else:
        return out
#--- End: def

def CF():
    '''Return the version of the CF conventions.

.. versionadded:: 1.7

:Returns:

    out:`str`
        The version of the CF conventions.

**Examples:**

>>> cfdm.CF()
'1.7'
    '''
    return __cf_version__
#--- End: def



from __future__ import print_function
import datetime
import os
import sys
import unittest

import numpy

import cfdm

verbose = False

class create_fieldTest(unittest.TestCase):

    def test_create_field(self):

        # Dimension coordinates
        dim1 = cfdm.core.DimensionCoordinate(data=cfdm.core.Data(cfdm.core.NumpyArray(numpy.arange(10.))))
        dim1.set_property('standard_name', 'grid_latitude')
        dim1.set_property('units', 'degrees')

        data = numpy.arange(9.) + 20
        data[-1] = 34
        dim0 = cfdm.core.DimensionCoordinate(data=cfdm.core.Data(cfdm.core.NumpyArray(data)))
        dim0.set_property('standard_name', 'grid_longitude')
        dim0.set_property('units', 'degrees')

        array = dim0.get_array()

        array = numpy.array([array-0.5, array+0.5]).transpose((1,0))
        array[-2, 1] = 30
        array[-1, :] = [30, 36]
        dim0.set_bounds(cfdm.core.Bounds(data=cfdm.core.Data(cfdm.core.NumpyArray(array))))
        
        dim2 = cfdm.core.DimensionCoordinate(
            data=cfdm.core.Data(cfdm.core.NumpyArray(numpy.array([1.5]))),
            bounds=cfdm.core.Bounds(data=cfdm.core.Data(cfdm.core.NumpyArray(numpy.array([[1, 2.]])))))
        dim2.set_property('standard_name'         , 'atmosphere_hybrid_height_coordinate')
        dim2.set_property('computed_standard_name', 'altitude')
                      
        # Auxiliary coordinates
        ak = cfdm.core.DomainAncillary(
            data=cfdm.core.Data(cfdm.core.NumpyArray(numpy.array([10.]))))
        ak.set_property('units', 'm')
        ak.set_bounds(cfdm.core.Bounds(
            data=cfdm.core.Data(cfdm.core.NumpyArray(numpy.array([[5, 15.]])))))
        
        bk = cfdm.core.DomainAncillary(
            data=cfdm.core.Data(cfdm.core.NumpyArray(numpy.array([20.]))))
        bk.set_bounds(cfdm.core.Bounds(
            data=cfdm.core.Data(cfdm.core.NumpyArray(numpy.array([[14, 26.]])))))
        
        aux2 = cfdm.core.AuxiliaryCoordinate(
            data=cfdm.core.Data(cfdm.core.NumpyArray(numpy.arange(-45, 45, dtype='int32').reshape(10, 9))))
        aux2.set_property('units', 'degree_N')
        aux2.set_property('standard_name', 'latitude')
        
        aux3 = cfdm.core.AuxiliaryCoordinate(
            data=cfdm.core.Data(cfdm.core.NumpyArray(numpy.arange(60, 150, dtype='int32').reshape(9, 10))))
        aux3.set_property('standard_name', 'longitude')
        aux3.set_property('units', 'degreeE')
        
        array = numpy.ma.array(['alpha','beta','gamma','delta','epsilon',
                                'zeta','eta','theta','iota','kappa'], dtype='S')
        array[0] = numpy.ma.masked
        aux4 = cfdm.core.AuxiliaryCoordinate(data=cfdm.core.Data(cfdm.core.NumpyArray(array)))
        aux4.set_property('long_name', 'greek_letters')
        
        # Cell measures
        msr0 = cfdm.core.CellMeasure(
            data=cfdm.core.Data(cfdm.core.NumpyArray(1+numpy.arange(90.).reshape(9, 10)*1234)))
        msr0.set_measure('area')
        msr0.set_property('units', 'km2')
        
        # Data          
        data = cfdm.core.Data(cfdm.core.NumpyArray(numpy.arange(90.).reshape(10, 9)))
        
        properties = {'units': 'm s-1'}
        
        f = cfdm.core.Field(properties=properties)
        f.set_property('standard_name', 'eastward_wind')

        axisX = f.set_construct(cfdm.core.DomainAxis(9))
        axisY = f.set_construct(cfdm.core.DomainAxis(10))
        axisZ = f.set_construct(cfdm.core.DomainAxis(1))

        f.set_data(data, axes=[axisY, axisX])
        
        x = f.set_construct(dim0, axes=[axisX])
        y = f.set_construct(dim1, axes=[axisY])
        z = f.set_construct(dim2, axes=[axisZ])

        lat   = f.set_construct(aux2, axes=[axisY, axisX])
        lon   = f.set_construct(aux3, axes=[axisX, axisY])
        greek = f.set_construct(aux4, axes=[axisY])

        ak = f.set_construct(ak, axes=[axisZ])
        bk = f.set_construct(bk, axes=[axisZ])

        # Coordinate references
        coordinate_conversion = cfdm.core.CoordinateConversion(
            parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                        'grid_north_pole_latitude': 38.0,
                        'grid_north_pole_longitude': 190.0})
        
        datum = cfdm.core.Datum(parameters={'earth_radius': 6371007})
        
        ref0 = cfdm.core.CoordinateReference(
            coordinate_conversion=coordinate_conversion,
            datum=datum,
            coordinates=[x, y, lat, lon]
        )

        f.set_construct(msr0, axes=[axisX, axisY])

        f.set_construct(ref0)

        orog = cfdm.core.DomainAncillary(data=f.get_data())
        orog.set_property('standard_name', 'surface_altitude')
        orog.set_property('units', 'm')
        orog = f.set_construct(orog, axes=[axisY, axisX])

        coordinate_conversion = cfdm.core.CoordinateConversion(
            parameters={'standard_name': 'atmosphere_hybrid_height_coordinate',
                        'computed_standard_name': 'altitude'},
            domain_ancillaries={'orog': orog,
                                'a'   : ak,
                                'b'   : bk})
        
        ref1 = cfdm.core.CoordinateReference(
            coordinates=[z],
            datum=datum,
            coordinate_conversion=coordinate_conversion
        )
        
        ref1 = f.set_construct(ref1)

        f_data = f.get_data()
#        print (repr(f_data), type(f_data))
#        print (repr(f_data.get_array()))
        # Field ancillary variables
        data = f_data
        anc = cfdm.core.FieldAncillary(data=data)
        f.set_construct(anc, axes=[axisY, axisX])
        
        data = f_data.get_array()[0]
        anc = cfdm.core.FieldAncillary(data=cfdm.core.Data(cfdm.core.NumpyArray(data)))
        f.set_construct(anc, axes=[axisX])

        data = f_data.get_array()[..., 0]
        anc = cfdm.core.FieldAncillary(data=cfdm.core.Data(cfdm.core.NumpyArray(data)))
        f.set_construct(anc, axes=[axisY])

        f.set_property('flag_values', numpy.array([1, 2, 4], 'int32'))
        f.set_property('flag_meanings', 'a bb ccc')
        f.set_property('flag_masks', [2, 1, 0])

        cm0 =  cfdm.core.CellMethod(
            axes=[axisX],
            properties={
                'method'   : 'mean',
                'intervals': [cfdm.core.Data(cfdm.core.NumpyArray(numpy.array(1)), 'day')],
                'comment'  : 'ok'})
        
        cm1 =  cfdm.core.CellMethod(
            axes=[axisY],
            properties={'method': 'maximum',
                        'where' : 'sea'})
        
        f.set_construct(cm0)
        f.set_construct(cm1)

#        print(f.get_data())
#        print(f.properties())
#        print(repr(f))
#        print(f)
#        print(f.constructs())
#        print(f.construct_axes())
    #--- End: def

#--- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cfdm.environment()
    print('')
    unittest.main(verbosity=2)



#from .. import __version__
from .. import CF

from .. import (CoordinateReference,
                Field,
                Data)

from . import CFDMImplementation

from .netcdf import NetCDFWrite


implementation = CFDMImplementation(cf_version=CF(),
                                    CoordinateReference=CoordinateReference,
                                    Field=Field,
                                    Data=Data)

def write(fields, filename, fmt='NETCDF4', overwrite=True,
          global_attributes=None, variable_attributes=None,
          external_file=None, datatype=None,
          least_significant_digit=None, endian='native', compress=0,
          fletcher32=False, shuffle=True, HDF_chunksizes=None,
          verbose=False, _implementation=implementation):
    '''Write field constructs to a netCDF file.

**File format**

See the *fmt* parameter for details on which output netCDF file
formats are supported.

**NetCDF variable and dimension names**

These names are stored within constructs read a from dataset, or may
be set manually. They are used when writing a field construct to the
file. If a name has not been set then one will be constructed (usually
based on the standard name if it exists). The names may be modified
internally to prevent duplication in the file.

Each construct, or construct component, that corresponds to a netCDF
variable has the following methods to get, set and remove a netCDF
variable name: `!nc_get_variable`, `!nc_set_variable` and
`!nc_del_variable` method

The domain axis construct has the following methods to get, set and
remove a netCDF dimension name: `~cfdm.DomainAxis.nc_get_dimension`,
`~cfdm.DomainAxis.nc_set_dimension` and
`~cfdm.DomainAxis.nc_del_dimension`.

**NetCDF attributes**

Field construct properties may be written as netCDF global attributes
or netCDF data variable attributes. See the *global_attributes* and
*variable_attributes* parameters for details.

**External variables**

Metadata constructs marked as external are omitted from the file and
referred to via the netCDF "external_variables" global
attribute. However, omitted constructs may be written to an external
file (see the *external_file* parameter for details). Constructs that
are allowed to be external have the `!nc_external` method to get and
set their external status:

**NetCDF unlimited dimensions**



**HDF chunksizes**



.. versionadded:: 1.7

.. seealso:: `cfdm.read`

:Parameters:

    fields: (sequence of) `Field`
        The field constructs to write to the file.

    filename: `str`
        The output netCDF file name. Various type of expansion are
        applied to the file names.

        Relative paths are allowed, and standard tilde and shell
        parameter expansions are applied to the string.

        *Example:*
          The file ``file.nc`` in the user's home directory could be
          described by any of the following: ``'$HOME/file.nc'``,
          ``'${HOME}/file.nc'``, ``'~/file.nc'``,
          ``'~/tmp/../file.nc'``.
  
    fmt: `str`, optional
        The format of the output file. One of:

          ==========================  =============================== 
          *fmt*                       Output file type                
          ==========================  =============================== 
          ``'NETCDF4'``               NetCDF4 format file. This is    
                                      the default.                    
                                                                      
          ``'NETCDF4_CLASSIC'``       NetCDF4 classic format file     
                                      (see below)                     
                                                                      
          ``'NETCDF3_CLASSIC'``       NetCDF3 classic format file 
                                      (limited to file sizes less     
                                      than 2GB).                      
                                                                      
          ``'NETCDF3_64BIT_OFFSET'``  NetCDF3 64-bit offset format    
                                      file                            
                                                                      
          ``'NETCDF3_64BIT'``         An alias for                    
                                      ``'NETCDF3_64BIT_OFFSET'``      
                                                                      
          ``'NETCDF3_64BIT_DATA'``    NetCDF3 64-bit offset format    
                                      file with extensions (see       
                                      below)                          
          ==========================  ================================

        By default the format is ``'NETCDF4'``.

        All formats support large files (i.e. those greater than 2GB)
        except ``'NETCDF3_CLASSIC'``.

        ``'NETCDF3_64BIT_DATA'`` is a format that requires version
        4.4.0 or newer of the C library (use `cfdm.environment` to see
        which version if the netCDF-C library is in use). It extends
        the ``'NETCDF3_64BIT_OFFSET'`` binary format to allow for
        unsigned/64 bit integer data types and 64-bit dimension sizes.

        ``'NETCDF4_CLASSIC'`` files use the version 4 disk format
        (HDF5), but omits features not found in the version 3
        API. They can be read by HDF5 clients. They can also be read
        by netCDF3 clients only if they have been re-linked against
        the netCDF4 library.

        ``'NETCDF4'`` files use the version 4 disk format (HDF5) and
        use the new features of the version 4 API.

    overwrite: `bool`, optional
        If False then raise an error if the output file pre-exists. By
        default a pre-existing output file is overwritten.

    global_attributes: (sequence of) `str`, optional
         Create netCDF global attributes from the specified field
         construct properties, rather than netCDF data variable
         attributes.

         These attributes are in addition to the following properties,
         which are created as netCDF global attributes by default:
         
           * the `description of file contents
             <http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.html#description-of-file-contents>`_
             properties, and

           * properties flagged as global on any of the field
             constructs being written (see
             `cfdm.Field.nc_global_attributes` for details).

         Note that it is not possible to create a netCDF global
         attribute from a property that has different values for
         different field constructs being written. In this case the
         property will not be written as a netCDF global attribute,
         even if it has been specified by the *global_attributes*
         parameter or is one of the default properties, but will
         appear as an attribute on the netCDF data variable
         corresponding to each field construct that contains the
         property.

         *Example:*
            ``global_attributes='project'``

         *Example:*
            ``global_attributes=['project', 'experiment']``

    variable_attributes: (sequence of) `str`, optional
         Create netCDF data variable attributes from the specified
         field construct properties.

         By default, all properties that are not created as netCDF
         global properties are created as attributes netCDF data
         variables. See the *global_attributes* parameter for details.

         Any property named by the *variable_attributes* parameter
         will always be created as a netCDF data variable attribute.

         *Example:*
            ``variable_attributes='project'``

         *Example:*
            ``variable_attributes=['project', 'doi']``

    external_file: `str`, optional   
        Write metadata constructs that have data and are marked as
        external to the named external file. Ignored if there are no
        such constructs.

    datatype: `dict`, optional
        Specify data type conversions to be applied prior to writing
        data to disk. This may be useful as a means of packing, or
        because the output format does not support a particular data
        type (for example, netCDF3 classic files do not support 64-bit
        integers). By default, input data types are preserved. Any
        data type conversion is only applied to the arrays on disk,
        and not to the input field constructs themselves.

        Data types conversions are defined by `numpy.dtype` objects in
        a dictionary whose keys are input data types with values of
        output data types.

        *Example:*

          To convert 64-bit integers to 32-bit integers:
          ``datatype={numpy.dtype('int64'): numpy.dtype('int32')}``.
       
    endian: `str`, optional
        The endian-ness of the output file. Valid values are
        ``'little'``, ``'big'`` or ``'native'``. By default the output
        is native endian. See the `netCDF4 package
        <http://unidata.github.io/netcdf4-python>`_ for more details.

        *Example:*
          ``endian='big'``

    compress: `int`, optional
        Regulate the speed and efficiency of compression. Must be an
        integer between ``0`` and ``9``. ``0`` means no compression;
        ``1`` is the fastest, but has the lowest compression ratio;
        ``9`` is the slowest but best compression ratio. The default
        value is ``0``. An error is raised if compression is requested
        for a netCDF3 output file format. See the `netCDF4 package
        <http://unidata.github.io/netcdf4-python>`_ for more details.

        *Example:*
          ``compress=4``
    
    least_significant_digit: `int`, optional
        Truncate the input field construct data arrays, but not the
        data arrays of metadata constructs. For a given positive
        integer, N the precision that is retained in the compressed
        data is 10 to the power -N. For example, a value of 2 will
        retain a precision of 0.01. In conjunction with the *compress*
        parameter this produces 'lossy', but significantly more
        efficient, compression. See the `netCDF4 package
        <http://unidata.github.io/netcdf4-python>`_ for more details.

        *Example:*
          ``least_significant_digit=3``

    fletcher32: `bool`, optional
        If True then the Fletcher-32 HDF5 checksum algorithm is
        activated to detect compression errors. Ignored if *compress*
        is ``0``. See the `netCDF4 package
        <http://unidata.github.io/netcdf4-python>`_ for details.

    shuffle: `bool`, optional
        If True (the default) then the HDF5 shuffle filter (which
        de-interlaces a block of data before compression by reordering
        the bytes by storing the first byte of all of a variable's
        values in the chunk contiguously, followed by all the second
        bytes, and so on) is turned off. By default the filter is
        applied because if the data array values are not all wildly
        different, using the filter can make the data more easily
        compressible.  Ignored if the *compress* parameter is ``0``
        (which is its default value). See the `netCDF4 package
        <http://unidata.github.io/netcdf4-python>`_ for more details.

    HDF_chunksizes: `dict`, optional

        Manually specify HDF5 chunks for the field construct data
        arrays.

        Chunking refers to a storage layout where a data array is
        partitioned into fixed-size multi-dimensional chunks when
        written to a netCDF4 file on disk. Chunking is ignored if the
        field construct is written to a netCDF3 format file.

        A chunk has the same rank as the data array, but with fewer
        (or no more) elements along each axes. The chunk is defined by
        a dictionary whose keys identify axes with values of the
        chunks size for those axes.

        If a given chunk size for an axis is larger than the axis size
        for any field, then the size of the axis at the time of
        writing to disk will be used instead.

        If chunk sizes have been specified for some but not all axes,
        then the each unset chunk size is assumed to be the full size
        of the axis for each field.

        If no chunk sizes have been set for any axes then the netCDF
        default chunk is used
        (http://www.unidata.ucar.edu/software/netcdf/docs/netcdf_perf_chunking.html).

        If any chunk sizes have already been set on a field construct
        with the `cf.Field.HDF_chunks` method then these are used in
        instead.

        A detailed discussion of HDF chunking and I/O performance is
        available at
        https://www.hdfgroup.org/HDF5/doc/H5.user/Chunking.html and
        http://www.unidata.ucar.edu/software/netcdf/workshops/2011/nc4chunking.
        Basically, you want the chunks for each dimension to match as
        closely as possible the size and shape of the data block that
        users will read from the file.

    verbose: `bool`, optional
        If True then print a summary of how constructs map to output
        netCDF dimensions, variables and attributes.

:Returns:

    `None`

**Examples:**

>>> cfdm.write(f, 'file.nc')
>>> cfdm.write(f, 'file.nc', fmt='NETCDF3_CLASSIC')
>>> cfdm.write(f, 'file.nc', external_file='cell_measures.nc')

    '''
    # ----------------------------------------------------------------
    # Initialise the netCDF write object
    # ----------------------------------------------------------------
    netcdf = NetCDFWrite(_implementation)

    if fields:
        netcdf.write(fields, filename, fmt=fmt, overwrite=overwrite,
                     global_attributes=global_attributes,
                     variable_attributes=variable_attributes,
                     external_file=external_file,
                     datatype=datatype,
                     least_significant_digit=least_significant_digit,
                     endian=endian, compress=compress,
                     shuffle=shuffle, fletcher32=fletcher32,
                     HDF_chunks=HDF_chunksizes, verbose=verbose)
#--- End: def



from builtins import (str, super)
from past.builtins import basestring

from copy import deepcopy

from . import abstract


class CellMethod(abstract.Properties):
    '''A cell method construct of the CF data model.

One or more cell method constructs describe how the cell values of the
field construct represent the variation of the physical quantity
within its cells, i.e. the structure of the data at a higher
resolution.

A single cell method construct consists of a set of axes, a "method"
property which describes how a value of the field construct's data
array describes the variation of the quantity within a cell over those
axes (e.g. a value might represent the cell area average), and
properties serving to indicate more precisely how the method was
applied (e.g. recording the spacing of the original data, or the fact
that the method was applied only over El Nino years).

.. versionadded:: 1.7
    '''
    def __init__(self, axes=None, properties=None, source=None,
                 copy=True):
        '''**Initialisation**

:Parameters:

    axes: sequence of `str`, optional
        Set the axes of the cell method construct. Either one or more
        domain axis construct identifiers or standard names. Ignored
        if the *source* parameter is set.

          *Example:*
             ``axes=['domainaxis0']``
        
          *Example:*
             ``axes=('domainaxis0', 'domainaxis1')``
        
          *Example:*
             ``axes=('area',)``
        
          *Example:*
             ``axes=['domainaxis2', 'time']``
        
         The axes may also be set after initialisation with the
         `set_axes` method.

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Example:*
          ``properties={'method': 'variance'}``
       
        *Example:*
          ``properties={'method': 'mean', 'where': 'sea'}``
        
        Properties may also be set after initialisation with the
        `properties` and `set_property` methods.

    source: optional
        Initialize the axes and properties from those of *source*.

    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization By default parameters are deep copied.

        '''
        super().__init__(properties=properties, source=source,
                         copy=copy)

        if source:
            try:
                axes = source.get_axes(None)
            except AttributeErrror:
                axes = None              
        #--- End: if

        if axes is not None:                
            axes = self.set_axes(axes)
    #--- End: def

#    def __repr__(self):
#        '''x.__repr__() <==> repr(x)
#
#        '''
#        return '<{0}: {1}>'.format(self.__class__.__name__, str(self))
#    #--- End: def
#
#    def __str__(self):
#        '''x.__str__() <==> str(x)
#
#Return a CF-netCDF-like string of the cell method.
#
#Note that if the intention is to use this string in a CF-netCDF
#cell_methods attribute then, unless they are standard names, the axes
#names will need to be modified to be netCDF dimension names.
#
#        '''     
#        string = ['{0}:'.format(axis) for axis in self.get_axes(())]
#
#        string.append(self.get_property('method', ''))
#
#        for portion in ('within', 'where', 'over'):
#            p = self.get_property(portion, None)
#            if p is not None:
#                string.extend((portion, p))
#        #--- End: for
#
#        intervals = self.get_property('intervals', ())
#        comment   = self.get_property('comment', None)
#
#        if intervals:
#            x = ['(']
#
#            y = ['interval: {0}'.format(data) for data in intervals]
#            x.append(' '.join(y))
#
#            if comment is not None:
#                x.append(' comment: {0}'.format(comment))
#
#            x.append(')')
#
#            string.append(''.join(x))
#
#        elif comment is not None:
#            string.append('({0})'.format(comment))
#
#        return ' '.join(string)
#    #--- End: def

    @property
    def construct_type(self):
        '''Return a description of the construct type.
        
.. versionadded:: 1.7
        
:Returns:

    out: `str`
        The construct type.

        '''
        return 'cell_method'

    def del_axes(self, *default):
        '''Remove the axes of the cell method.

.. versionadded:: 1.7

.. seealso:: `get_axes`, `has_axes`, `set_axes`

:Parameters:

    default: optional
        Return *default* if axes have not been set.

:Returns:

     out:
        The removed axes, identified by field key or standard name. If
        unset then *default* is returned, if provided.

**Examples**

>>> c.set_axes('area')
>>> print c.del_axes()
('area',)
>>> print f.del_axes()
None

        '''
        return self._del_component('axes')
    #--- End: def
    
    def get_axes(self, *default):
        '''Return the axes of the cell method.

.. versionadded:: 1.7

.. seealso:: `del_axes`, `has_axes`, `set_axes`

:Parameters:

    default: optional
        Return *default* if and only if axes have not been set.

:Returns:

    out: `tuple`
        The axes. If axes have not been set then return the value
        of *default* parameter, if provided.

**Examples**

>>> c.set_axes(['time'])
>>> c.get_axes()
('time',)
>>> c.del_axes()
>>> c.get_axes()
AttributeError: 'CellMethod' object has no component 'axes'
>>> c.get_axes('NO AXES')
'NO AXES'

        '''
        return self._get_component('axes', *default)
    #--- End: def

    def has_axes(self):
        '''Whether the axes of the cell method have been set.

.. versionadded:: 1.7

.. seealso:: `del_axes`, `get_axes`, `set_axes`

:Examples 1:

>>> x = c.has_axes()

:Returns:

     out: `bool`
        True if the axes have been set, otherwise False.

**Examples**

>>> if c.has_axes():
...     print 'Has axes'

'''
        return self._has_component('axes')
    #--- End: def

    def set_axes(self, value, copy=True):
        '''Set the axes of the cell method.

.. versionadded:: 1.7

.. seealso:: `del_axes`, `get_axes`, `has_axes`

:Examples 1:

>>> c.set_axes('time')

:Parameters:

    value: (sequence of) `str`
        The value for the axes.

:Returns:

     `None`

        '''
        if copy:
            value = deepcopy(value)
        
        if isinstance(value, basestring):
            value = (value,)
        else:
            value = tuple(value)
            
        return self._set_component('axes', value, copy=False)
    #--- End: def

#--- End: class



from . import abstract


class DimensionCoordinate(abstract.Coordinate):
    '''A dimension coordinate construct of the CF data model.

A dimension coordinate construct provides information which locate the
cells of the domain and which depend on a subset of the domain axis
constructs. The dimension coordinate construct is able to
unambiguously describe cell locations because a domain axis can be
associated with at most one dimension coordinate construct, whose data
array values must all be non-missing and strictly monotonically
increasing or decreasing. They must also all be of the same numeric
data type. If cell bounds are provided, then each cell must have
exactly two vertices. CF-netCDF coordinate variables and numeric
scalar coordinate variables correspond to dimension coordinate
constructs.

The dimension coordinate construct consists of a data array of the
coordinate values which spans a subset of the domain axis constructs,
an optional array of cell bounds recording the extents of each cell
(stored in a `Bounds` object), and properties to describe the
coordinates. An array of cell bounds spans the same domain axes as its
coordinate array, with the addition of an extra dimension whose size
is that of the number of vertices of each cell. This extra dimension
does not correspond to a domain axis construct since it does not
relate to an independent axis of the domain. Note that, for
climatological time axes, the bounds are interpreted in a special way
indicated by the cell method constructs.

    '''
#    def get_construct_type(self):


    @property
    def construct_type(self):
        '''Return a description of the construct type.
        
.. versionadded:: 1.7
        
:Returns:

    out: `str`
        The construct type.

        '''
        return 'dimension_coordinate'
    #--- End: def

#--- End: class



'''cfdm is a complete implementation of the CF data model, that
identifies the fundamental elements of the CF conventions and shows
how they relate to each other, independently of the netCDF encoding.

The central element defined by the CF data model is the field
construct, which corresponds to a CF-netCDF data variable with all of
its metadata.

The cfdm package can

    * read field constructs from netCDF datasets,
    * create new field constructs in memory,
    * inspect field constructs,
    * test whether two field constructs are the same,
    * modify field construct metadata and data,
    * create subspaces of field constructs,
    * write field constructs to netCDF datasets on disk,
    * incorporate, and create, metadata stored in external files, and
    * read, write, and create data that have been compressed by
      convention (i.e. ragged or gathered arrays), whilst presenting a
      view of the data in its uncompressed form.

'''

__author__       = 'David Hassell'
__date__         = '2018-12-02'
__cf_version__   = '1.7'
__version__      = '1.7b8'

from distutils.version import StrictVersion
import platform

# Check the version of python
if StrictVersion(platform.python_version()) < StrictVersion('2.7.0'):
    raise ValueError(
        "Bad python version: cfdm requires python 2.7 or later. Got {}".format(
        platform.python_version()))

from .constants        import *
from .constructs       import Constructs

from .data import (Data,
                   Array,
                   CompressedArray,
                   NumpyArray,
                   NetCDFArray,
                   GatheredArray,
                   RaggedContiguousArray,
                   RaggedIndexedArray,
                   RaggedIndexedContiguousArray)

from .functions import (CF, environment, ATOL, RTOL)

from .count import Count
from .index import Index
from .list  import List

from .bounds               import Bounds
from .coordinateconversion import CoordinateConversion
from .datum                import Datum
from .domain               import Domain
from .interiorring         import InteriorRing

from .auxiliarycoordinate import AuxiliaryCoordinate
from .cellmeasure         import CellMeasure
from .cellmethod          import CellMethod
from .coordinatereference import CoordinateReference
from .dimensioncoordinate import DimensionCoordinate
from .domainancillary     import DomainAncillary
from .domainaxis          import DomainAxis
from .field               import Field
from .fieldancillary      import FieldAncillary

from .io import (read, write)


    
